# TODO: use tx_payer_is_user to simplify accounting
# TODO: implement "blank" V3 pools and incorporate try/except for all V3 pool manager get_pool calls
# TODO: add state block argument for pool simulation calls
# TODO: instead of appending pool states to list, replace with dict and only return final state

import contextlib
from typing import TYPE_CHECKING, Any, Self

import eth_abi.abi
from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3

from degenbot import connection_manager, get_checksum_address
from degenbot.constants import WRAPPED_NATIVE_TOKENS
from degenbot.erc20 import Erc20Token, Erc20TokenManager
from degenbot.exceptions import DegenbotError, DegenbotValueError
from degenbot.exceptions.evm import EVMRevertError
from degenbot.exceptions.liquidity_pool import LiquidityPoolError
from degenbot.exceptions.manager import ManagerError
from degenbot.exceptions.transaction import (
    DeadlineExpired,
    InsufficientInput,
    InsufficientOutput,
    LeftoverRouterBalance,
    PreviousBlockMismatch,
    TransactionError,
    UnknownRouterAddress,
)
from degenbot.logging import logger
from degenbot.transaction.simulation_ledger import SimulationLedger
from degenbot.types.abstract import AbstractSimulationResult, AbstractTransaction
from degenbot.types.aliases import BlockNumber, ChainId
from degenbot.uniswap.abi import UNISWAP_V3_ROUTER2_ABI, UNISWAP_V3_ROUTER_ABI
from degenbot.uniswap.deployments import (
    ROUTER_DEPLOYMENTS,
    UniswapRouterDeployment,
    UniswapV2ExchangeDeployment,
    UniswapV3ExchangeDeployment,
)
from degenbot.uniswap.managers import UniswapV2PoolManager, UniswapV3PoolManager
from degenbot.uniswap.v2_functions import generate_v2_pool_address, get_v2_pools_from_token_path
from degenbot.uniswap.v2_liquidity_pool import UniswapV2Pool, UnregisteredLiquidityPool
from degenbot.uniswap.v2_types import UniswapV2PoolSimulationResult, UniswapV2PoolState
from degenbot.uniswap.v3_functions import decode_v3_path
from degenbot.uniswap.v3_liquidity_pool import UniswapV3Pool
from degenbot.uniswap.v3_types import UniswapV3PoolSimulationResult, UniswapV3PoolState


class UniversalRouterSpecialAddress:
    # ref: https://github.com/Uniswap/universal-router/blob/deployed-commit/contracts/libraries/Constants.sol
    ETH = get_checksum_address("0x0000000000000000000000000000000000000000")
    MSG_SENDER = get_checksum_address("0x0000000000000000000000000000000000000001")
    ROUTER = get_checksum_address("0x0000000000000000000000000000000000000002")


class UniversalRouterSpecialValues:
    # ref: https://github.com/Uniswap/universal-router/blob/deployed-commit/contracts/libraries/Constants.sol
    V2_PAIR_ALREADY_PAID = 0
    USE_CONTRACT_BALANCE = 1 << 255


class V3RouterSpecialAddress:
    # SwapRouter.sol checks for address(0)
    # ref: https://github.com/Uniswap/v3-periphery/blob/main/contracts/SwapRouter.sol
    ROUTER_1 = get_checksum_address("0x0000000000000000000000000000000000000000")

    # ref: https://github.com/Uniswap/swap-router-contracts/blob/main/contracts/libraries/Constants.sol
    MSG_SENDER = get_checksum_address("0x0000000000000000000000000000000000000001")
    ROUTER_2 = get_checksum_address("0x0000000000000000000000000000000000000002")


class V3RouterSpecialValues:
    # ref: https://github.com/Uniswap/swap-router-contracts/blob/main/contracts/libraries/Constants.sol
    USE_CONTRACT_BALANCE = 0


class UniswapTransaction(AbstractTransaction):
    @classmethod
    def from_router(cls, router: UniswapRouterDeployment, **kwargs: Any) -> Self:
        return cls(
            chain_id=router.chain_id,
            router_address=router.address,
            **kwargs,
        )

    def __init__(
        self,
        chain_id: ChainId | str,
        router_address: str,
        func_name: str,
        func_params: dict[str, Any],
        tx_hash: HexBytes | bytes | str,
        tx_nonce: int | str,
        tx_value: int | str,
        tx_sender: str,
    ) -> None:
        """
        Build a standalone representation of a transaction submitted to a known Uniswap-based
        router contract.

        Supported contracts:
            Uniswap V2 Router
            Uniswap V2 Router 2
            Uniswap V3 Router
            Uniswap V3 Router 2
            Uniswap Universal Router
        """

        # @dev The `self.ledger` is used to track token balances for all
        # addresses involved in the swap. A positive balance represents
        # a pre-swap deposit, a negative balance represents an outstanding
        # withdrawal. A full transaction should end with a positive balance of
        # the desired output token, credited to `self.sender` or one or more
        # recipients, which are collected in the `self.to` set.

        # @dev Some routers have special flags that signify the swap amount
        # should be looked up at the time of the swap, as opposed to a
        # specified amount at the time the transaction is built. The ledger is
        # used to look up the balance at any point inside the swap, and at the
        # end to confirm that all balances have been accounted for.

        self.ledger = SimulationLedger()
        self.chain_id: ChainId

        self.v2_pool_manager: UniswapV2PoolManager | None = None
        self.v3_pool_manager: UniswapV3PoolManager | None = None

        self.chain_id = int(chain_id, 16) if isinstance(chain_id, str) else chain_id
        self.router_address = get_checksum_address(router_address)
        if self.router_address not in ROUTER_DEPLOYMENTS[self.chain_id]:
            raise UnknownRouterAddress

        router_deployment = ROUTER_DEPLOYMENTS[self.chain_id][self.router_address]

        # Create pool managers for the supported exchanges
        for exchange in router_deployment.exchanges:
            match exchange:
                case UniswapV2ExchangeDeployment():
                    self.v2_pool_manager = UniswapV2PoolManager.get_instance(
                        factory_address=exchange.factory.address, chain_id=self.chain_id
                    ) or UniswapV2PoolManager(
                        factory_address=exchange.factory.address, chain_id=self.chain_id
                    )
                case UniswapV3ExchangeDeployment():
                    self.v3_pool_manager = UniswapV3PoolManager.get_instance(
                        factory_address=exchange.factory.address, chain_id=self.chain_id
                    ) or UniswapV3PoolManager(
                        factory_address=exchange.factory.address, chain_id=self.chain_id
                    )
                case _:
                    raise DegenbotValueError(message=f"Could not identify DEX type for {exchange}")

        self.sender = get_checksum_address(tx_sender)
        self.recipients: set[ChecksumAddress] = set()

        self.hash = HexBytes(tx_hash)
        self.nonce = int(tx_nonce, 16) if isinstance(tx_nonce, str) else tx_nonce
        self.value = int(tx_value, 16) if isinstance(tx_value, str) else tx_value
        self.func_name = func_name
        self.func_params = func_params
        if previous_block_hash := self.func_params.get("previousBlockhash"):
            self.func_previous_block_hash = HexBytes(previous_block_hash)

        self.silent = False

    def _raise_if_past_deadline(
        self, deadline: int, block_number: BlockNumber
    ) -> None:  # pragma: no cover
        block = connection_manager.get_web3(self.chain_id).eth.get_block(block_number)
        block_timestamp = block.get("timestamp")
        if block_timestamp is not None and block_timestamp > deadline:
            raise DeadlineExpired

    def _raise_if_block_hash_mismatch(self, block_hash: HexBytes) -> None:  # pragma: no cover
        logger.info(f"Checking previousBlockhash: {block_hash!r}")
        block = connection_manager.get_web3(self.chain_id).eth.get_block("latest")
        _block_hash = block.get("hash")
        if _block_hash is not None and block_hash != _block_hash:
            raise PreviousBlockMismatch

    @staticmethod
    def _show_pool_states(
        pool: UniswapV2Pool | UniswapV3Pool,
        sim_result: UniswapV2PoolSimulationResult | UniswapV3PoolSimulationResult,
    ) -> None:
        current_state = sim_result.initial_state
        future_state = sim_result.final_state

        # amount out is negative
        if sim_result.amount0_delta < sim_result.amount1_delta:
            token_in = pool.token1
            token_out = pool.token0
            amount_in = sim_result.amount1_delta
            amount_out = -sim_result.amount0_delta
        else:
            token_in = pool.token0
            token_out = pool.token1
            amount_in = sim_result.amount0_delta
            amount_out = -sim_result.amount1_delta

        logger.info(f"Simulating swap through pool: {pool}")
        logger.info(f"\t{amount_in} {token_in} -> {amount_out} {token_out}")
        match current_state, future_state, sim_result:
            case UniswapV2PoolState(), UniswapV2PoolState(), UniswapV2PoolSimulationResult():
                logger.info("\t(CURRENT)")
                logger.info(f"\t{pool.token0}: {current_state.reserves_token0}")
                logger.info(f"\t{pool.token1}: {current_state.reserves_token1}")
                logger.info("\t(FUTURE)")
                logger.info(f"\t{pool.token0}: {future_state.reserves_token0}")
                logger.info(f"\t{pool.token1}: {future_state.reserves_token1}")
            case UniswapV3PoolState(), UniswapV3PoolState(), UniswapV3PoolSimulationResult():
                logger.info(f"\t{amount_in} {token_in} -> {amount_out} {token_out}")
                logger.info("\t(CURRENT)")
                logger.info(f"\tprice={current_state.sqrt_price_x96}")
                logger.info(f"\tliquidity={current_state.liquidity}")
                logger.info(f"\ttick={current_state.tick}")
                logger.info("\t(FUTURE)")
                logger.info(f"\tprice={future_state.sqrt_price_x96}")
                logger.info(f"\tliquidity={future_state.liquidity}")
                logger.info(f"\ttick={future_state.tick}")

    def _simulate_v2_swap_exact_in(
        self,
        *,
        pool: UniswapV2Pool,
        recipient: ChecksumAddress | str,
        token_in: Erc20Token,
        amount_in: int,
        amount_out_min: int | None = None,
        first_swap: bool = False,
        last_swap: bool = False,
    ) -> tuple[
        UniswapV2Pool,
        UniswapV2PoolSimulationResult,
    ]:
        assert isinstance(pool, UniswapV2Pool), f"Called _simulate_v2_swap_exact_in on pool {pool}"

        token_out = pool.token1 if token_in == pool.token0 else pool.token0

        if amount_in in (
            UniversalRouterSpecialValues.USE_CONTRACT_BALANCE,
            V3RouterSpecialValues.USE_CONTRACT_BALANCE,
        ):
            _balance = self.ledger.token_balance(self.router_address, token_in)

            self.ledger.transfer(
                token=token_in,
                amount=_balance,
                to_addr=pool.address,
                from_addr=self.router_address,
            )
            amount_in = self.ledger.token_balance(pool.address, token_in)

        sim_result = pool.simulate_exact_input_swap(
            token_in=token_in,
            token_in_quantity=amount_in,
        )

        _amount_out = -min(
            sim_result.amount0_delta,
            sim_result.amount1_delta,
        )

        if first_swap and not self.ledger.token_balance(pool.address, token_in):
            # credit the router if there is a zero balance
            if not self.ledger.token_balance(self.router_address, token_in):
                self.ledger.adjust(self.router_address, token_in.address, amount_in)
            self.ledger.transfer(
                token_in,
                amount=amount_in,
                from_addr=self.router_address,
                to_addr=pool.address,
            )

        # process the swap
        self.ledger.adjust(pool.address, token_in.address, -amount_in)
        self.ledger.adjust(pool.address, token_out.address, _amount_out)

        # transfer to the recipient
        self.ledger.transfer(
            token=token_out,
            amount=_amount_out,
            from_addr=pool.address,
            to_addr=recipient,
        )

        if last_swap:
            self.recipients.add(get_checksum_address(recipient))

        if last_swap and amount_out_min is not None and _amount_out < amount_out_min:
            raise InsufficientOutput(
                minimum=amount_out_min,
                received=_amount_out,
            )

        if not self.silent:
            self._show_pool_states(pool, sim_result)

        return pool, sim_result

    @staticmethod
    def _simulate_v2_add_liquidity(
        pool: UniswapV2Pool,
        added_reserves_token0: int,
        added_reserves_token1: int,
        override_state: UniswapV2PoolState | None = None,
    ) -> UniswapV2PoolSimulationResult:
        return pool.simulate_add_liquidity(
            added_reserves_token0=added_reserves_token0,
            added_reserves_token1=added_reserves_token1,
            override_state=override_state,
        )

    def _simulate_v2_swap_exact_out(
        self,
        *,
        pool: UniswapV2Pool,
        recipient: ChecksumAddress | str,
        token_in: Erc20Token,
        amount_out: int,
        amount_in_max: int | None = None,
        first_swap: bool = False,
    ) -> tuple[UniswapV2Pool, UniswapV2PoolSimulationResult]:
        assert isinstance(pool, UniswapV2Pool), f"Called _simulate_v2_swap_exact_out on pool {pool}"

        token_out = pool.token1 if token_in == pool.token0 else pool.token0

        sim_result = pool.simulate_exact_output_swap(
            token_out=token_out,
            token_out_quantity=amount_out,
        )

        _amount_in = max(
            sim_result.amount0_delta,
            sim_result.amount1_delta,
        )

        if first_swap:
            # transfer the input token amount from the sender to the first pool
            self.ledger.transfer(
                token=token_in.address,
                amount=_amount_in,
                from_addr=self.sender,
                to_addr=pool.address,
            )

        # process the swap
        self.ledger.adjust(pool.address, token_in.address, -_amount_in)
        self.ledger.adjust(pool.address, token_out.address, amount_out)

        # transfer the output token from the pool to the recipient
        self.ledger.transfer(
            token=token_out.address,
            amount=amount_out,
            from_addr=pool.address,
            to_addr=recipient,
        )

        if first_swap and amount_in_max is not None and _amount_in > amount_in_max:
            raise InsufficientInput(minimum=_amount_in, deposited=amount_in_max)

        if not self.silent:
            self._show_pool_states(pool, sim_result)

        return pool, sim_result

    def _simulate_v3_swap_exact_in(
        self,
        *,
        pool: UniswapV3Pool,
        recipient: str,
        token_in: Erc20Token,
        amount_in: int,
        amount_out_min: int | None = None,
        sqrt_price_limit_x96: int | None = None,
        first_swap: bool = False,
    ) -> tuple[UniswapV3Pool, UniswapV3PoolSimulationResult]:
        assert isinstance(pool, UniswapV3Pool), f"Called _simulate_v3_swap_exact_in on pool {pool}"

        self.recipients.add(get_checksum_address(recipient))

        token_out = pool.token1 if token_in == pool.token0 else pool.token0

        if amount_in in (
            UniversalRouterSpecialValues.USE_CONTRACT_BALANCE,
            V3RouterSpecialValues.USE_CONTRACT_BALANCE,
        ):
            amount_in = self.ledger.token_balance(self.router_address, token_in.address)

        # the swap may occur after wrapping ETH, in which case amountIn will
        # be already set. If not, credit the router (user transfers the input)
        if first_swap and not self.ledger.token_balance(self.router_address, token_in.address):
            self.ledger.adjust(
                self.router_address,
                token_in.address,
                amount_in,
            )

        try:
            _sim_result = pool.simulate_exact_input_swap(
                token_in=token_in,
                token_in_quantity=amount_in,
                sqrt_price_limit_x96=sqrt_price_limit_x96,
            )
        except EVMRevertError as exc:
            raise TransactionError(message=f"V3 revert: {exc}") from exc

        _amount_out = -min(
            _sim_result.amount0_delta,
            _sim_result.amount1_delta,
        )

        self.ledger.adjust(
            self.router_address,
            token_in.address,
            -amount_in,
        )

        logger.debug(f"{recipient=}")
        match recipient:
            case UniversalRouterSpecialAddress.MSG_SENDER:
                recipient = self.sender
            case (
                UniversalRouterSpecialAddress.ROUTER
                | V3RouterSpecialAddress.ROUTER_1
                | V3RouterSpecialAddress.ROUTER_2
            ):
                recipient = self.router_address
            case _:
                pass

        self.ledger.adjust(
            recipient,
            token_out.address,
            _amount_out,
        )

        if amount_out_min is not None and _amount_out < amount_out_min:
            raise InsufficientOutput(minimum=amount_out_min, received=_amount_out)

        if not self.silent:
            self._show_pool_states(pool, _sim_result)

        return pool, _sim_result

    def _simulate_v3_swap_exact_out(
        self,
        *,
        pool: UniswapV3Pool,
        recipient: str,
        token_in: Erc20Token,
        amount_out: int,
        amount_in_max: int | None = None,
        sqrt_price_limit_x96: int | None = None,
        first_swap: bool = False,
        last_swap: bool = False,
    ) -> tuple[UniswapV3Pool, UniswapV3PoolSimulationResult]:
        assert isinstance(pool, UniswapV3Pool), f"Called _simulate_v3_swap_exact_out on pool {pool}"

        self.recipients.add(get_checksum_address(recipient))

        token_out = pool.token1 if token_in == pool.token0 else pool.token0

        try:
            _sim_result = pool.simulate_exact_output_swap(
                token_out=token_out,
                token_out_quantity=amount_out,
                sqrt_price_limit_x96=sqrt_price_limit_x96,
            )
        except EVMRevertError as exc:
            raise TransactionError(message=f"V3 revert: {exc}") from exc

        _amount_in = max(
            _sim_result.amount0_delta,
            _sim_result.amount1_delta,
        )
        _amount_out = -min(
            _sim_result.amount0_delta,
            _sim_result.amount1_delta,
        )

        # adjust the post-swap balances for each token
        self.ledger.adjust(
            self.router_address,
            token_in,
            -_amount_in,
        )
        self.ledger.adjust(
            self.router_address,
            token_out,
            _amount_out,
        )

        if first_swap:
            # Exact output swaps proceed in reverse order, so the last
            # iteration will show a negative balance of the input token,
            # which must be accounted for.

            router_input_token_balance = self.ledger.token_balance(self.router_address, token_in)

            # Check for a balance:
            #   - If zero or positive, take no action
            #   - If negative, adjust with the assumption the user has paid
            #     that amount with the transaction call
            if router_input_token_balance < 0:
                self.ledger.adjust(
                    self.router_address,
                    token_in,
                    _amount_in,
                )

            if amount_in_max is not None and amount_in_max < _amount_in:
                raise InsufficientInput(minimum=_amount_in, deposited=amount_in_max)

        if last_swap:
            match recipient:
                case UniversalRouterSpecialAddress.MSG_SENDER:
                    _recipient = self.sender
                case (
                    UniversalRouterSpecialAddress.ROUTER
                    | V3RouterSpecialAddress.ROUTER_1
                    | V3RouterSpecialAddress.ROUTER_2
                ):
                    _recipient = self.router_address
                case _:
                    _recipient = get_checksum_address(recipient)
                    self.recipients.add(_recipient)

            self.ledger.transfer(
                token=token_out.address,
                amount=_amount_out,
                from_addr=self.router_address,
                to_addr=_recipient,
            )

        if not self.silent:
            self._show_pool_states(pool, _sim_result)

        return pool, _sim_result

    def _simulate_unwrap(self, wrapped_token: str) -> None:
        logger.debug(f"Unwrapping {wrapped_token}")

        wrapped_token_balance = self.ledger.token_balance(self.router_address, wrapped_token)

        self.ledger.adjust(
            self.router_address,
            wrapped_token,
            -wrapped_token_balance,
        )

    def _simulate_sweep(self, token: str, recipient: str) -> None:
        logger.debug(f"Sweeping {token} to {recipient}")

        token_balance = self.ledger.token_balance(self.router_address, token)
        self.ledger.adjust(
            self.router_address,
            token,
            -token_balance,
        )
        self.ledger.adjust(
            recipient,
            token,
            token_balance,
        )

    def _simulate(
        self,
        func_name: str,
        func_params: dict[str, Any],
        block_number: BlockNumber,
    ) -> None:
        """
        Take a Uniswap V2 / V3 transaction (specified by name and a dictionary
        of arguments to that function) and return a list of pools and state
        dictionaries for all pools used by the transaction
        """

        v2_functions = {
            "addLiquidity",
            "addLiquidityETH",
            "swapExactTokensForETH",
            "swapExactTokensForETHSupportingFeeOnTransferTokens",
            "swapExactETHForTokens",
            "swapExactETHForTokensSupportingFeeOnTransferTokens",
            "swapExactTokensForTokens",
            "swapExactTokensForTokensSupportingFeeOnTransferTokens",
            "swapTokensForExactETH",
            "swapTokensForExactTokens",
            "swapETHForExactTokens",
        }

        v3_functions = {
            "exactInputSingle",
            "exactInput",
            "exactOutputSingle",
            "exactOutput",
            "multicall",
            "sweepToken",
            "unwrapWETH9",
            "unwrapWETH9WithFee",
            "wrapETH",
        }

        universal_router_functions = {
            "execute",
        }

        unhandled_functions = {
            # TODO: handle these
            "increaseLiquidity",
            "removeLiquidity",
            "removeLiquidityETH",
            "removeLiquidityETHWithPermit",
            "removeLiquidityETHSupportingFeeOnTransferTokens",
            "removeLiquidityETHWithPermitSupportingFeeOnTransferTokens",
            "removeLiquidityWithPermit",
            "sweepTokenWithFee",
            # V3 multicall functions
            # ref: https://github.com/Uniswap/swap-router-contracts/blob/main/contracts/base/ApproveAndCall.sol
            "mint",
        }

        no_op_functions = {
            # These functions do not affect the pool state.
            # ---
            # ref: https://docs.uniswap.org/contracts/v3/reference/periphery/interfaces/IPeripheryPayments#refundeth
            "refundETH",
            # ---
            #
            # EIP-2612 token permit functions
            # ref: https://docs.uniswap.org/contracts/v3/reference/periphery/base/SelfPermit
            "selfPermit",
            "selfPermitAllowed",
            "selfPermitAllowedIfNecessary",
            "selfPermitIfNecessary",
            # ---
            # ref: https://github.com/Uniswap/swap-router-contracts/blob/main/contracts/base/PeripheryPaymentsExtended.sol
            "pull",
        }

        def _process_universal_router_command(
            command_type: int,
            inputs: bytes,
        ) -> None:
            # ref: https://github.com/Uniswap/universal-router/blob/main/contracts/libraries/Commands.sol
            universal_router_command_values: dict[int, str | None] = {
                0x00: "V3_SWAP_EXACT_IN",
                0x01: "V3_SWAP_EXACT_OUT",
                0x02: "PERMIT2_TRANSFER_FROM",
                0x03: "PERMIT2_PERMIT_BATCH",
                0x04: "SWEEP",
                0x05: "TRANSFER",
                0x06: "PAY_PORTION",
                0x07: None,  # COMMAND_PLACEHOLDER
                0x08: "V2_SWAP_EXACT_IN",
                0x09: "V2_SWAP_EXACT_OUT",
                0x0A: "PERMIT2_PERMIT",
                0x0B: "WRAP_ETH",
                0x0C: "UNWRAP_WETH",
                0x0D: "PERMIT2_TRANSFER_FROM_BATCH",
                0x0E: "BALANCE_CHECK_ERC20",
                0x0F: None,  # COMMAND_PLACEHOLDER
                0x10: "SEAPORT_V1_5",
                0x11: "LOOKS_RARE_V2",
                0x12: "NFTX",
                0x13: "CRYPTOPUNKS",
                0x14: "LOOKS_RARE_1155",  # dropped in V1_3
                0x15: "OWNER_CHECK_721",
                0x16: "OWNER_CHECK_1155",
                0x17: "SWEEP_ERC721",
                0x18: "X2Y2_721",
                0x19: "SUDOSWAP",
                0x1A: "NFT20",
                0x1B: "X2Y2_1155",
                0x1C: "FOUNDATION",
                0x1D: "SWEEP_ERC1155",
                0x1E: "ELEMENT_MARKET",
                0x1F: None,  # COMMAND_PLACEHOLDER
                0x20: "SEAPORT_V1_4",
                0x21: "EXECUTE_SUB_PLAN",
                0x22: "APPROVE_ERC20",
                0x23: None,  # COMMAND_PLACEHOLDER
                0x24: None,  # COMMAND_PLACEHOLDER
                0x25: None,  # COMMAND_PLACEHOLDER
                0x26: None,  # COMMAND_PLACEHOLDER
                0x27: None,  # COMMAND_PLACEHOLDER
                0x28: None,  # COMMAND_PLACEHOLDER
                0x29: None,  # COMMAND_PLACEHOLDER
                0x2A: None,  # COMMAND_PLACEHOLDER
                0x2B: None,  # COMMAND_PLACEHOLDER
                0x2C: None,  # COMMAND_PLACEHOLDER
                0x2D: None,  # COMMAND_PLACEHOLDER
                0x2E: None,  # COMMAND_PLACEHOLDER
                0x2F: None,  # COMMAND_PLACEHOLDER
                0x30: None,  # COMMAND_PLACEHOLDER
                0x31: None,  # COMMAND_PLACEHOLDER
                0x32: None,  # COMMAND_PLACEHOLDER
                0x33: None,  # COMMAND_PLACEHOLDER
                0x34: None,  # COMMAND_PLACEHOLDER
                0x35: None,  # COMMAND_PLACEHOLDER
                0x36: None,  # COMMAND_PLACEHOLDER
                0x37: None,  # COMMAND_PLACEHOLDER
                0x38: None,  # COMMAND_PLACEHOLDER
                0x39: None,  # COMMAND_PLACEHOLDER
                0x3A: None,  # COMMAND_PLACEHOLDER
                0x3B: None,  # COMMAND_PLACEHOLDER
                0x3C: None,  # COMMAND_PLACEHOLDER
                0x3D: None,  # COMMAND_PLACEHOLDER
                0x3E: None,  # COMMAND_PLACEHOLDER
                0x3F: None,  # COMMAND_PLACEHOLDER
            }

            unimplemented_universal_router_commands = {
                "APPROVE_ERC20",
                "BALANCE_CHECK_ERC20",
                "CRYPTOPUNKS",
                "ELEMENT_MARKET",
                "EXECUTE_SUB_PLAN",
                "FOUNDATION",
                "LOOKS_RARE_1155",
                "LOOKS_RARE_721",
                "NFT20",
                "NFTX",
                "OWNER_CHECK_1155",
                "OWNER_CHECK_721",
                "PERMIT2_PERMIT_BATCH",
                "PERMIT2_PERMIT",
                "PERMIT2_TRANSFER_FROM_BATCH",
                "PERMIT2_TRANSFER_FROM",
                "SEAPORT_V2",
                "SEAPORT",
                "SUDOSWAP",
                "SWEEP_ERC1155",
                "SWEEP_ERC721",
                "X2Y2_1155",
                "X2Y2_721",
            }

            command_type_mask = 0x3F
            command = universal_router_command_values[command_type & command_type_mask]

            logger.debug(f"Processing Universal Router command: {command}")

            if TYPE_CHECKING:
                _amount_in: int = 0
                _amount_out: int = 0
                _sim_result: UniswapV2PoolSimulationResult | UniswapV3PoolSimulationResult

            match command:
                case "SWEEP":
                    """
                    This function transfers the current token balance held by the contract to
                    `recipient`
                    """

                    sweep_token_address, sweep_recipient, sweep_amount_min = eth_abi.abi.decode(
                        types=("address", "address", "uint256"),
                        data=inputs,
                    )

                    sweep_recipient = get_checksum_address(sweep_recipient)
                    match sweep_recipient:  # pragma: no cover
                        case UniversalRouterSpecialAddress.MSG_SENDER:
                            sweep_recipient = self.sender
                        case UniversalRouterSpecialAddress.ROUTER:
                            sweep_recipient = self.router_address
                        case _:
                            pass

                    sweep_token_balance = self.ledger.token_balance(
                        self.router_address, sweep_token_address
                    )

                    if sweep_token_balance < sweep_amount_min:
                        raise InsufficientOutput(
                            minimum=sweep_amount_min,
                            received=sweep_token_balance,
                        )

                    self._simulate_sweep(sweep_token_address, sweep_recipient)

                case "PAY_PORTION":
                    """
                    Transfers a portion of the current token balance held by the
                    contract to `recipient`
                    """

                    _pay_portion_token_address, _pay_portion_recipient, _pay_portion_bips = (
                        eth_abi.abi.decode(
                            types=("address", "address", "uint256"),
                            data=inputs,
                        )
                    )

                    # ref: https://docs.uniswap.org/contracts/universal-router/technical-reference#pay_portion
                    # ref: https://github.com/Uniswap/universal-router/blob/main/contracts/libraries/Constants.sol
                    # TODO: refactor if ledger needs to support ETH balances
                    if _pay_portion_token_address == UniversalRouterSpecialAddress.ETH:
                        logger.info("PAY_PORTION called with ETH shorthand")

                    else:
                        sweep_token_balance = self.ledger.token_balance(
                            self.router_address, _pay_portion_token_address
                        )
                        self.ledger.transfer(
                            _pay_portion_token_address,
                            sweep_token_balance * _pay_portion_bips // 10_000,
                            self.router_address,
                            _pay_portion_recipient,
                        )
                        self.recipients.add(get_checksum_address(_pay_portion_recipient))

                case "TRANSFER":
                    """
                    Transfer an `amount` of `token` to `recipient`
                    """

                    transfer_token, transfer_recipient, transfer_value = eth_abi.abi.decode(
                        types=("address", "address", "uint256"), data=inputs
                    )
                    transfer_recipient = get_checksum_address(transfer_recipient)
                    self.ledger.adjust(
                        address=self.router_address, token=transfer_token, amount=-transfer_value
                    )
                    self.ledger.adjust(
                        address=transfer_recipient, token=transfer_token, amount=transfer_value
                    )

                case "WRAP_ETH":
                    """
                    This function wraps a quantity of ETH to WETH and transfers it
                    to `recipient`.

                    The mainnet WETH contract only implements the `deposit` method,
                    so `recipient` will always be the router address.

                    Some L2s and side chains implement a `depositTo` method, so
                    `recipient` is evaluated before adjusting the ledger balance.
                    """

                    tx_recipient: ChecksumAddress

                    _tx_recipient, _wrap_amount_min = eth_abi.abi.decode(
                        types=("address", "uint256"),
                        data=inputs,
                    )

                    _recipient = get_checksum_address(_tx_recipient)
                    match _tx_recipient:  # pragma: no cover
                        case UniversalRouterSpecialAddress.MSG_SENDER:
                            _recipient = self.sender
                        case UniversalRouterSpecialAddress.ROUTER:
                            _recipient = self.router_address
                        case _:
                            pass

                    self.ledger.adjust(
                        address=_recipient,
                        token=WRAPPED_NATIVE_TOKENS[self.chain_id],
                        amount=_wrap_amount_min,
                    )

                case "UNWRAP_WETH":
                    """
                    Unwraps a quantity of WETH held by the router to ETH.
                    """

                    _, _unwrap_amount_min = eth_abi.abi.decode(
                        types=("address", "uint256"),
                        data=inputs,
                    )

                    _wrapped_token_address = WRAPPED_NATIVE_TOKENS[self.chain_id]
                    _wrapped_token_balance = self.ledger.token_balance(
                        self.router_address, _wrapped_token_address
                    )

                    if _wrapped_token_balance < _unwrap_amount_min:
                        raise InsufficientOutput(
                            minimum=_unwrap_amount_min,
                            received=_wrapped_token_balance,
                        )

                    self._simulate_unwrap(_wrapped_token_address)

                case "V2_SWAP_EXACT_IN":
                    """
                    Decode an exact input swap through Uniswap V2 liquidity pools.

                    Returns: a list of tuples representing the pool object and the
                    final state of the pool after the swap completes.
                    """

                    tx_recipient, tx_amount_in, tx_amount_out_min, tx_path, tx_payer_is_user = (
                        eth_abi.abi.decode(
                            types=(
                                "address",
                                "uint256",
                                "uint256",
                                "address[]",
                                "bool",
                            ),
                            data=inputs,
                        )
                    )

                    try:
                        if TYPE_CHECKING:
                            assert self.v2_pool_manager is not None
                        pools = get_v2_pools_from_token_path(tx_path, self.v2_pool_manager)
                    except (LiquidityPoolError, ManagerError):
                        raise TransactionError(
                            message=f"Pools could not be built for all steps in path {tx_path}"
                        ) from None

                    last_pool_pos = len(tx_path) - 2

                    for pool_pos, pool in enumerate(pools):
                        first_swap = pool_pos == 0
                        last_swap = pool_pos == last_pool_pos

                        token_in = pool.token0 if tx_path[pool_pos] == pool.token0 else pool.token1

                        _amount_in = tx_amount_in if first_swap else _amount_out

                        _recipient = tx_recipient if last_swap else pools[pool_pos + 1].address

                        match _recipient:
                            case UniversalRouterSpecialAddress.MSG_SENDER:
                                _recipient = self.sender
                            case (
                                UniversalRouterSpecialAddress.ROUTER
                                | V3RouterSpecialAddress.ROUTER_1
                                | V3RouterSpecialAddress.ROUTER_2
                            ):
                                _recipient = self.router_address

                        _, _sim_result = self._simulate_v2_swap_exact_in(
                            pool=pool,
                            recipient=_recipient,
                            token_in=token_in,
                            amount_in=_amount_in,
                            amount_out_min=tx_amount_out_min if last_swap else None,
                            first_swap=first_swap,
                            last_swap=last_swap,
                        )

                        _amount_out = -min(_sim_result.amount0_delta, _sim_result.amount1_delta)

                        self.simulated_pool_states.append((pool, _sim_result))

                case "V2_SWAP_EXACT_OUT":
                    """
                    Decode an exact output swap through Uniswap V2 liquidity pools.

                    Returns: a list of tuples representing the pool object and the
                    final state of the pool after the swap completes.
                    """

                    tx_recipient, tx_amount_out, tx_amount_in_max, tx_path, tx_payer_is_user = (
                        eth_abi.abi.decode(
                            types=(
                                "address",
                                "uint256",
                                "uint256",
                                "address[]",
                                "bool",
                            ),
                            data=inputs,
                        )
                    )

                    match tx_recipient:
                        case UniversalRouterSpecialAddress.ROUTER:
                            tx_recipient = self.router_address
                        case UniversalRouterSpecialAddress.MSG_SENDER:
                            tx_recipient = self.sender
                        case _:
                            pass

                    try:
                        if TYPE_CHECKING:
                            assert self.v2_pool_manager is not None
                        pools = get_v2_pools_from_token_path(tx_path, self.v2_pool_manager)
                    except (LiquidityPoolError, ManagerError):
                        raise TransactionError(
                            message=f"Pools could not be built for all steps in path {tx_path}"
                        ) from None

                    last_pool_pos = len(pools) - 1

                    for pool_pos, pool in enumerate(pools[::-1]):
                        first_swap = pool_pos == last_pool_pos
                        last_swap = pool_pos == 0

                        _amount_out = tx_amount_out if last_swap else _amount_in
                        _amount_in_max = tx_amount_in_max if first_swap else None

                        _recipient = (
                            tx_recipient if last_swap else pools[::-1][pool_pos - 1].address
                        )

                        # get the input token by proceeding backwards from
                        # the 2nd to last position,
                        # e.g. for a token0 -> token1 -> token2 -> token3
                        # swap proceeding through pool0 -> pool1 -> pool2,
                        # - the last swap (pool_pos = 0) will have the input token
                        #   in the second to last position (index -2, token2)
                        # - the middle swap (pool_pos = 1) will have input token
                        #   in the third to last position (index -3, token1)
                        # - the first swap (pool_pos = 2) will have input token
                        #   in the four to last position (index -4, token0)
                        _token_in = (
                            pool.token0 if pool.token0 == tx_path[-2 - pool_pos] else pool.token1
                        )

                        _, _v2_sim_result = self._simulate_v2_swap_exact_out(
                            pool=pool,
                            recipient=_recipient,
                            token_in=_token_in,
                            amount_out=_amount_out,
                            amount_in_max=_amount_in_max,
                            first_swap=first_swap,
                        )

                        _amount_in = max(
                            _v2_sim_result.amount0_delta,
                            _v2_sim_result.amount1_delta,
                        )

                        self.simulated_pool_states.append((pool, _v2_sim_result))

                case "V3_SWAP_EXACT_IN":
                    """
                    Decode an exact input swap through Uniswap V3 liquidity pools.

                    Returns: a list of tuples representing the pool object and the final state of
                    the pool after the swap completes.
                    """

                    tx_recipient, tx_amount_in, tx_amount_out_min, tx_path, tx_payer_is_user = (
                        eth_abi.abi.decode(
                            types=(
                                "address",
                                "uint256",
                                "uint256",
                                "bytes",
                                "bool",
                            ),
                            data=inputs,
                        )
                    )

                    tx_path_decoded = decode_v3_path(tx_path)

                    # decode the path - tokenIn is the first position, fee is the
                    # second position, tokenOut is the third position.
                    # paths can be an arbitrary length, but address & fee values
                    # are always interleaved. e.g. tokenIn, fee, tokenOut, fee,
                    last_token_pos = len(tx_path_decoded) - 3

                    for token_pos in range(
                        0,
                        len(tx_path_decoded) - 2,
                        2,
                    ):
                        tx_token_in_address = tx_path_decoded[token_pos]
                        fee = tx_path_decoded[token_pos + 1]
                        tx_token_out_address = tx_path_decoded[token_pos + 2]

                        if TYPE_CHECKING:
                            assert isinstance(tx_token_in_address, str)
                            assert isinstance(fee, int)
                            assert isinstance(tx_token_out_address, str)

                        first_swap = token_pos == 0
                        last_swap = token_pos == last_token_pos

                        if TYPE_CHECKING:
                            assert self.v3_pool_manager is not None
                        v3_pool = self.v3_pool_manager.get_pool_from_tokens_and_fee(
                            token_addresses=(
                                tx_token_in_address,
                                tx_token_out_address,
                            ),
                            pool_fee=fee,
                            silent=self.silent,
                        )

                        _recipient = (
                            tx_recipient if last_swap else UniversalRouterSpecialAddress.ROUTER
                        )
                        _token_in = (
                            v3_pool.token0
                            if v3_pool.token0 == tx_token_in_address
                            else v3_pool.token1
                        )
                        _amount_in = tx_amount_in if first_swap else _amount_out
                        _amount_out_min = tx_amount_out_min if last_swap else None

                        v3_sim_result: UniswapV3PoolSimulationResult
                        _, v3_sim_result = self._simulate_v3_swap_exact_in(
                            pool=v3_pool,
                            recipient=_recipient,
                            token_in=_token_in,
                            amount_in=_amount_in,
                            amount_out_min=_amount_out_min,
                            first_swap=first_swap,
                        )

                        _amount_out = -min(
                            v3_sim_result.amount0_delta,
                            v3_sim_result.amount1_delta,
                        )

                        self.simulated_pool_states.append((v3_pool, v3_sim_result))

                case "V3_SWAP_EXACT_OUT":
                    """
                    Decode an exact output swap through Uniswap V3 liquidity pools.

                    Returns: a list of tuples representing the pool object and the
                    final state of the pool after the swap completes.
                    """

                    tx_recipient, tx_amount_out, tx_amount_in_max, tx_path, tx_payer_is_user = (
                        eth_abi.abi.decode(
                            types=(
                                "address",
                                "uint256",
                                "uint256",
                                "bytes",
                                "bool",
                            ),
                            data=inputs,
                        )
                    )

                    tx_path_decoded = decode_v3_path(tx_path)

                    # An exact output path is encoded in REVERSE order,
                    # tokenOut is the first position, tokenIn is the second
                    # position. e.g. tokenOut, fee, tokenIn
                    last_token_pos = len(tx_path_decoded) - 3

                    for token_pos in range(
                        0,
                        len(tx_path_decoded) - 2,
                        2,
                    ):
                        tx_token_out_address = tx_path_decoded[token_pos]
                        fee = tx_path_decoded[token_pos + 1]
                        tx_token_in_address = tx_path_decoded[token_pos + 2]

                        if TYPE_CHECKING:
                            assert isinstance(tx_token_in_address, str)
                            assert isinstance(tx_token_out_address, str)
                            assert isinstance(fee, int)

                        first_swap = token_pos == last_token_pos
                        last_swap = token_pos == 0

                        if TYPE_CHECKING:
                            assert self.v3_pool_manager is not None
                        v3_pool = self.v3_pool_manager.get_pool_from_tokens_and_fee(
                            token_addresses=(
                                tx_token_in_address,
                                tx_token_out_address,
                            ),
                            pool_fee=fee,
                            silent=self.silent,
                        )

                        _recipient = (
                            tx_recipient if last_swap else UniversalRouterSpecialAddress.ROUTER
                        )
                        _token_in = (
                            v3_pool.token0
                            if v3_pool.token0 == tx_token_in_address
                            else v3_pool.token1
                        )
                        _amount_out = tx_amount_out if last_swap else _amount_in
                        _amount_in_max = tx_amount_in_max if first_swap else None

                        _, v3_sim_result = self._simulate_v3_swap_exact_out(
                            pool=v3_pool,
                            recipient=_recipient,
                            token_in=_token_in,
                            amount_out=_amount_out,
                            amount_in_max=_amount_in_max,
                            first_swap=first_swap,
                            last_swap=last_swap,
                        )

                        _amount_in = max(
                            v3_sim_result.amount0_delta,
                            v3_sim_result.amount1_delta,
                        )

                        # check that the output of each intermediate swap meets
                        # the input for the next swap
                        if not last_swap:
                            # pool states are appended to `future_pool_states`
                            # so the previous swap will be in the last position
                            _, _last_sim_result = self.simulated_pool_states[-1]

                            if TYPE_CHECKING:
                                assert isinstance(
                                    _last_sim_result,
                                    UniswapV2PoolSimulationResult | UniswapV2PoolSimulationResult,
                                )

                            _last_amount_in = max(
                                _last_sim_result.amount1_delta,
                                _last_sim_result.amount0_delta,
                            )

                            if _amount_out != _last_amount_in:
                                raise TransactionError(
                                    message=f"Insufficient swap amount through requested pool {v3_pool}. Needed {_last_amount_in}, received {_amount_out}"  # noqa:E501
                                )

                        self.simulated_pool_states.append((v3_pool, v3_sim_result))

                case _:
                    if command in unimplemented_universal_router_commands:
                        logger.debug(f"UNIMPLEMENTED COMMAND: {command}")
                    else:  # pragma: no cover
                        raise DegenbotValueError(message=f"Invalid command {command}")

        def _process_v3_multicall(
            params: dict[str, Any],
        ) -> None:
            with contextlib.suppress(KeyError):
                self._raise_if_block_hash_mismatch(params["previousBlockhash"])

            for payload in params["data"]:
                payload_func = payload_args = None
                with contextlib.suppress(ValueError):
                    # decode with Router ABI
                    payload_func, payload_args = (
                        Web3()
                        .eth.contract(abi=UNISWAP_V3_ROUTER_ABI)
                        .decode_function_input(payload)
                    )

                with contextlib.suppress(ValueError):
                    # decode with Router2 ABI
                    payload_func, payload_args = (
                        Web3()
                        .eth.contract(abi=UNISWAP_V3_ROUTER2_ABI)
                        .decode_function_input(payload)
                    )

                if payload_func is None or payload_args is None:  # pragma: no cover
                    raise DegenbotValueError(message="Failed to decode payload.")

                # special case to handle a multicall encoded within
                # another multicall
                if payload_func.fn_name == "multicall":
                    logger.debug("Unwrapping nested multicall")

                    with contextlib.suppress(KeyError):
                        self._raise_if_block_hash_mismatch(params["previousBlockhash"])

                    for function_input in payload_args["data"]:
                        _func = _params = None
                        with contextlib.suppress(ValueError):
                            _func, _params = (
                                Web3()
                                .eth.contract(abi=UNISWAP_V3_ROUTER_ABI)
                                .decode_function_input(function_input)
                            )

                        with contextlib.suppress(ValueError):
                            _func, _params = (
                                Web3()
                                .eth.contract(abi=UNISWAP_V3_ROUTER2_ABI)
                                .decode_function_input(function_input)
                            )

                        if _func is None or _params is None:  # pragma: no cover
                            raise DegenbotValueError(
                                message="Failed to decode function parameters."
                            )

                        self._simulate(
                            func_name=_func.fn_name,
                            func_params=_params,
                            block_number=block_number,
                        )

                else:
                    self._simulate(
                        func_name=payload_func.fn_name,
                        func_params=payload_args,
                        block_number=block_number,
                    )

        def _process_uniswap_v2_transaction() -> None:
            if TYPE_CHECKING:
                _amount_in: int = 0
                _amount_out: int = 0
                assert self.v2_pool_manager is not None

            try:
                match func_name:
                    case (
                        "swapExactTokensForETH"
                        | "swapExactTokensForETHSupportingFeeOnTransferTokens"
                        | "swapExactETHForTokens"
                        | "swapExactETHForTokensSupportingFeeOnTransferTokens"
                        | "swapExactTokensForTokens"
                        | "swapExactTokensForTokensSupportingFeeOnTransferTokens"
                    ):
                        logger.debug(f"{func_name}: {self.hash.to_0x_hex()=}")

                        try:
                            tx_amount_in = func_params["amountIn"]
                        except KeyError:
                            tx_amount_in = self.value  # 'swapExactETHForTokens'

                        tx_amount_out_min = func_params["amountOutMin"]
                        tx_path = func_params["path"]

                        tx_recipient = func_params["to"]
                        if tx_recipient == UniversalRouterSpecialAddress.ROUTER:
                            tx_recipient = self.router_address

                        try:
                            tx_deadline = func_params["deadline"]
                        except KeyError:
                            pass
                        else:
                            self._raise_if_past_deadline(
                                tx_deadline,
                                block_number=block_number,
                            )

                        try:
                            pools = get_v2_pools_from_token_path(tx_path, self.v2_pool_manager)
                        except (LiquidityPoolError, ManagerError):
                            raise TransactionError(
                                message=f"Pools could not be built for all steps in path {tx_path}"
                            ) from None

                        last_pool_pos = len(tx_path) - 2

                        for pool_pos, pool in enumerate(pools):
                            first_swap = pool_pos == 0
                            last_swap = pool_pos == last_pool_pos

                            _token_in = (
                                pool.token0 if tx_path[pool_pos] == pool.token0 else pool.token1
                            )

                            _amount_in = tx_amount_in if first_swap else _amount_out

                            _recipient = tx_recipient if last_swap else pools[pool_pos + 1].address

                            _, sim_result = self._simulate_v2_swap_exact_in(
                                pool=pool,
                                recipient=_recipient,
                                token_in=_token_in,
                                amount_in=_amount_in,
                                amount_out_min=tx_amount_out_min if last_swap else None,
                                first_swap=first_swap,
                                last_swap=last_swap,
                            )

                            _amount_out = -min(sim_result.amount0_delta, sim_result.amount1_delta)

                            self.simulated_pool_states.append((pool, sim_result))

                    case (
                        "swapTokensForExactETH"
                        | "swapTokensForExactTokens"
                        | "swapETHForExactTokens"
                    ):
                        logger.debug(f"{func_name}: {self.hash.to_0x_hex()=}")

                        tx_amount_out = func_params["amountOut"]
                        try:
                            tx_amount_in_max = func_params["amountInMax"]
                        except KeyError:
                            tx_amount_in_max = self.value  # 'swapETHForExactTokens'
                        tx_path = func_params["path"]

                        tx_recipient = func_params["to"]
                        if tx_recipient == UniversalRouterSpecialAddress.ROUTER:
                            tx_recipient = self.router_address
                        else:
                            self.recipients.add(tx_recipient)

                        with contextlib.suppress(KeyError):
                            tx_deadline = func_params["deadline"]
                            self._raise_if_past_deadline(
                                tx_deadline,
                                block_number=block_number,
                            )

                        try:
                            pools = get_v2_pools_from_token_path(tx_path, self.v2_pool_manager)
                        except (LiquidityPoolError, ManagerError) as exc:  # pragma: no cover
                            raise TransactionError(
                                message=f"Pools could not be built for all steps in path {tx_path}"
                            ) from exc

                        last_pool_pos = len(pools) - 1

                        for pool_pos, pool in enumerate(pools[::-1]):
                            first_swap = pool_pos == last_pool_pos
                            last_swap = pool_pos == 0

                            _amount_out = tx_amount_out if last_swap else _amount_in
                            _amount_in_max = tx_amount_in_max if first_swap else None

                            _recipient = (
                                tx_recipient if last_swap else pools[::-1][pool_pos - 1].address
                            )

                            # get the input token by proceeding backwards from
                            # the 2nd to last position,
                            # e.g. for a token0 -> token1 -> token2 -> token3
                            # swap proceeding through pool0 -> pool1 -> pool2,
                            # - the last swap (pool_pos = 0) will have the input token
                            #   in the second to last position (index -2, token2)
                            # - the middle swap (pool_pos = 1) will have input token
                            #   in the third to last position (index -3, token1)
                            # - the first swap (pool_pos = 2) will have input token
                            #   in the four to last position (index -4, token0)
                            _token_in = (
                                pool.token0
                                if pool.token0 == tx_path[-2 - pool_pos]
                                else pool.token1
                            )

                            _, _sim_result = self._simulate_v2_swap_exact_out(
                                pool=pool,
                                recipient=_recipient,
                                token_in=_token_in,
                                amount_out=_amount_out,
                                amount_in_max=_amount_in_max,
                                first_swap=first_swap,
                            )

                            _amount_in = max(
                                _sim_result.amount0_delta,
                                _sim_result.amount1_delta,
                            )

                            if first_swap:
                                self.ledger.adjust(
                                    self.sender,
                                    WRAPPED_NATIVE_TOKENS[self.chain_id],
                                    _amount_in,
                                )

                            self.simulated_pool_states.append((pool, _sim_result))

                    case "addLiquidity" | "addLiquidityETH":
                        logger.debug(f"{func_name}: {self.hash.to_0x_hex()=}")

                        if func_name == "addLiquidity":
                            tx_token_a = get_checksum_address(func_params["tokenA"])
                            tx_token_b = get_checksum_address(func_params["tokenB"])
                            tx_token_amount_a = func_params["amountADesired"]
                            tx_token_amount_b = func_params["amountBDesired"]
                            token0_address, token1_address = (
                                (tx_token_a, tx_token_b)
                                if tx_token_a < tx_token_b
                                else (tx_token_b, tx_token_a)
                            )
                            token0_amount, token1_amount = (
                                (tx_token_amount_a, tx_token_amount_b)
                                if tx_token_a < tx_token_b
                                else (tx_token_amount_b, tx_token_amount_a)
                            )
                        else:
                            tx_token = get_checksum_address(func_params["token"])
                            tx_token_amount = func_params["amountTokenDesired"]
                            tx_eth_min = func_params["amountETHMin"]
                            _wrapped_token_address = WRAPPED_NATIVE_TOKENS[self.chain_id]
                            token0_address, token1_address = (
                                (_wrapped_token_address, tx_token)
                                if _wrapped_token_address < tx_token
                                else (tx_token, _wrapped_token_address)
                            )
                            token0_amount, token1_amount = (
                                (tx_eth_min, tx_token_amount)
                                if _wrapped_token_address < tx_token
                                else (tx_token_amount, tx_eth_min)
                            )

                        tx_deadline = func_params["deadline"]
                        self._raise_if_past_deadline(
                            tx_deadline,
                            block_number=block_number,
                        )

                        try:
                            _pool = self.v2_pool_manager.get_pool_from_tokens(
                                token_addresses=(token0_address, token1_address),
                                silent=self.silent,
                            )
                        except ManagerError:
                            token_manager = Erc20TokenManager(chain_id=self.chain_id)
                            _pool = UnregisteredLiquidityPool(
                                address=generate_v2_pool_address(
                                    token_addresses=(
                                        token0_address,
                                        token1_address,
                                    ),
                                    deployer_address=self.v2_pool_manager.factory_address,
                                    init_hash=self.v2_pool_manager.pool_init_hash,
                                ),
                                tokens=[
                                    token_manager.get_erc20token(
                                        token0_address, silent=self.silent
                                    ),
                                    token_manager.get_erc20token(
                                        token1_address, silent=self.silent
                                    ),
                                ],
                            )

                        if TYPE_CHECKING:
                            assert isinstance(_pool, UniswapV2Pool)
                        _sim_result = self._simulate_v2_add_liquidity(
                            pool=_pool,
                            added_reserves_token0=token0_amount,
                            added_reserves_token1=token1_amount,
                            override_state=UniswapV2PoolState(
                                address=_pool.address,
                                reserves_token0=0,
                                reserves_token1=0,
                                block=None,
                            ),
                        )

                        self.simulated_pool_states.append((_pool, _sim_result))

                    case _:
                        raise DegenbotValueError(message=f"Unknown function: {func_name}!")

            except TransactionError:
                # Catch specific subclass exception to prevent nested
                # multicalls from being recursively re-annotated
                # e.g. 'Simulation failed: Simulation failed: {error}'
                raise
            except DegenbotError as e:
                raise TransactionError(message=f"Simulation failed: {e}") from e

        def _process_uniswap_v3_transaction() -> None:
            logger.debug(f"{func_name}: {self.hash.to_0x_hex()=}")

            if TYPE_CHECKING:
                v3_sim_result: UniswapV3PoolSimulationResult
                assert self.v3_pool_manager is not None

            try:
                match func_name:
                    case "multicall":
                        _process_v3_multicall(params=func_params)

                    case "exactInputSingle":
                        match func_params["params"], len(func_params["params"]):
                            case dict(), 7 | 8:
                                tx_token_in_address = func_params["params"]["tokenIn"]
                                tx_token_out_address = func_params["params"]["tokenOut"]
                                tx_fee = func_params["params"]["fee"]
                                tx_recipient = func_params["params"]["recipient"]
                                tx_amount_in = func_params["params"]["amountIn"]
                                tx_amount_out_min = func_params["params"]["amountOutMinimum"]
                                tx_deadline = func_params["params"].get("deadline")
                                tx_sqrt_price_limit_x96 = func_params["params"]["sqrtPriceLimitX96"]
                            case tuple(), 8:
                                # Decode with ISwapRouter ABI
                                # ref: https://github.com/Uniswap/v3-periphery/blob/main/contracts/interfaces/ISwapRouter.sol
                                (
                                    tx_token_in_address,
                                    tx_token_out_address,
                                    tx_fee,
                                    tx_recipient,
                                    tx_deadline,
                                    tx_amount_in,
                                    tx_amount_out_min,
                                    tx_sqrt_price_limit_x96,
                                ) = func_params["params"]
                            case tuple(), 7:
                                # Decode with IV3SwapRouter ABI (aka Router2)
                                # ref: https://github.com/Uniswap/swap-router-contracts/blob/main/contracts/interfaces/IV3SwapRouter.sol
                                (
                                    tx_token_in_address,
                                    tx_token_out_address,
                                    tx_fee,
                                    tx_recipient,
                                    tx_amount_in,
                                    tx_amount_out_min,
                                    tx_sqrt_price_limit_x96,
                                ) = func_params["params"]
                                tx_deadline = None
                            case _:
                                raise DegenbotValueError(
                                    message=f"Could not identify function parameters. Got {(func_params['params'])}"  # noqa:E501
                                )

                        if tx_deadline:
                            self._raise_if_past_deadline(
                                tx_deadline,
                                block_number=block_number,
                            )

                        if TYPE_CHECKING:
                            assert isinstance(self.v3_pool_manager, UniswapV3PoolManager)
                        v3_pool = self.v3_pool_manager.get_pool_from_tokens_and_fee(
                            token_addresses=(
                                tx_token_in_address,
                                tx_token_out_address,
                            ),
                            pool_fee=tx_fee,
                            silent=self.silent,
                        )

                        _, v3_sim_result = self._simulate_v3_swap_exact_in(
                            pool=v3_pool,
                            recipient=tx_recipient,
                            token_in=(
                                v3_pool.token0
                                if v3_pool.token0 == tx_token_in_address
                                else v3_pool.token1
                            ),
                            amount_in=tx_amount_in,
                            amount_out_min=tx_amount_out_min,
                            first_swap=True,
                        )

                        self.simulated_pool_states.append((v3_pool, v3_sim_result))

                        token_out_quantity = -min(
                            v3_sim_result.amount1_delta,
                            v3_sim_result.amount0_delta,
                        )

                    case "exactInput":
                        match func_params["params"], len(func_params["params"]):
                            case dict(), 4 | 5:
                                tx_path = func_params["params"]["path"]
                                tx_recipient = func_params["params"]["recipient"]
                                tx_deadline = func_params["params"].get("deadline")
                                tx_amount_in = func_params["params"]["amountIn"]
                                tx_amount_out_minimum = func_params["params"]["amountOutMinimum"]
                            case tuple(), 5:
                                # Decode with ISwapRouter ABI
                                # ref: https://github.com/Uniswap/v3-periphery/blob/main/contracts/interfaces/ISwapRouter.sol
                                (
                                    tx_path,
                                    tx_recipient,
                                    tx_deadline,
                                    tx_amount_in,
                                    tx_amount_out_minimum,
                                ) = func_params["params"]
                            case tuple(), 4:
                                # Decode with IV3SwapRouter ABI (aka Router2)
                                # ref: https://github.com/Uniswap/swap-router-contracts/blob/main/contracts/interfaces/IV3SwapRouter.sol
                                (
                                    tx_path,
                                    tx_recipient,
                                    tx_amount_in,
                                    tx_amount_out_minimum,
                                ) = func_params["params"]
                                tx_deadline = None
                            case _:
                                raise DegenbotValueError(
                                    message=f"Could not identify function parameters. Got {(func_params['params'])}"  # noqa:E501
                                )

                        if tx_deadline:
                            self._raise_if_past_deadline(
                                tx_deadline,
                                block_number=block_number,
                            )

                        tx_path_decoded = decode_v3_path(tx_path)

                        if not self.silent:
                            logger.info(f" • path = {tx_path_decoded}")
                            logger.info(f" • recipient = {tx_recipient}")
                            if tx_deadline is not None:
                                logger.info(f" • deadline = {tx_deadline}")
                            logger.info(f" • amountIn = {tx_amount_in}")
                            logger.info(f" • amountOutMinimum = {tx_amount_out_minimum}")

                        last_token_pos = len(tx_path_decoded) - 3
                        token_out_quantity = 0  # this is overridden by the first calc

                        for token_pos in range(
                            0,
                            len(tx_path_decoded) - 2,
                            2,
                        ):
                            tx_token_in_address = tx_path_decoded[token_pos]
                            tx_fee = tx_path_decoded[token_pos + 1]
                            tx_token_out_address = tx_path_decoded[token_pos + 2]

                            if TYPE_CHECKING:
                                assert isinstance(tx_token_in_address, str)
                                assert isinstance(tx_fee, int)
                                assert isinstance(tx_token_out_address, str)

                            first_swap = token_pos == 0
                            last_swap = token_pos == last_token_pos

                            v3_pool = self.v3_pool_manager.get_pool_from_tokens_and_fee(
                                token_addresses=(
                                    tx_token_in_address,
                                    tx_token_out_address,
                                ),
                                pool_fee=tx_fee,
                                silent=self.silent,
                            )

                            _, v3_sim_result = self._simulate_v3_swap_exact_in(
                                pool=v3_pool,
                                recipient=tx_recipient
                                if last_swap
                                else UniversalRouterSpecialAddress.ROUTER,
                                token_in=v3_pool.token0
                                if v3_pool.token0 == tx_token_in_address
                                else v3_pool.token1,
                                amount_in=tx_amount_in if token_pos == 0 else token_out_quantity,
                                # only apply minimum output to the last swap
                                amount_out_min=tx_amount_out_minimum if last_swap else None,
                                first_swap=first_swap,
                            )

                            self.simulated_pool_states.append((v3_pool, v3_sim_result))

                            token_out_quantity = -min(
                                v3_sim_result.amount0_delta,
                                v3_sim_result.amount1_delta,
                            )

                    case "exactOutputSingle":
                        match func_params["params"], len(func_params["params"]):
                            case dict(), 7 | 8:
                                tx_token_in_address = func_params["params"]["tokenIn"]
                                tx_token_out_address = func_params["params"]["tokenOut"]
                                tx_fee = func_params["params"]["fee"]
                                tx_recipient = func_params["params"]["recipient"]
                                tx_deadline = func_params["params"].get("deadline")
                                tx_amount_out = func_params["params"]["amountOut"]
                                tx_amount_in_max = func_params["params"]["amountInMaximum"]
                                tx_sqrt_price_limit_x96 = func_params["params"]["sqrtPriceLimitX96"]
                            case tuple(), 8:
                                # Decode with ISwapRouter ABI
                                # https://github.com/Uniswap/v3-periphery/blob/main/contracts/interfaces/ISwapRouter.sol
                                (
                                    tx_token_in_address,
                                    tx_token_out_address,
                                    tx_fee,
                                    tx_recipient,
                                    tx_deadline,
                                    tx_amount_out,
                                    tx_amount_in_max,
                                    tx_sqrt_price_limit_x96,
                                ) = func_params["params"]

                            case tuple(), 7:
                                # Decode with IV3SwapRouter ABI (aka Router2)
                                # https://github.com/Uniswap/swap-router-contracts/blob/main/contracts/interfaces/IV3SwapRouter.sol
                                (
                                    tx_token_in_address,
                                    tx_token_out_address,
                                    tx_fee,
                                    tx_recipient,
                                    tx_amount_out,
                                    tx_amount_in_max,
                                    tx_sqrt_price_limit_x96,
                                ) = func_params["params"]
                                tx_deadline = None
                            case _:
                                raise DegenbotValueError(
                                    message=f"Could not identify function parameters. Got {(func_params['params'])}"  # noqa:E501
                                )

                        if tx_deadline:
                            self._raise_if_past_deadline(
                                tx_deadline,
                                block_number=block_number,
                            )

                        if TYPE_CHECKING:
                            assert isinstance(self.v3_pool_manager, UniswapV3PoolManager)
                        v3_pool = self.v3_pool_manager.get_pool_from_tokens_and_fee(
                            token_addresses=(
                                tx_token_in_address,
                                tx_token_out_address,
                            ),
                            pool_fee=tx_fee,
                            silent=self.silent,
                        )

                        _, v3_sim_result = self._simulate_v3_swap_exact_out(
                            pool=v3_pool,
                            recipient=tx_recipient,
                            token_in=v3_pool.token0
                            if v3_pool.token0 == tx_token_in_address
                            else v3_pool.token1,
                            amount_out=tx_amount_out,
                            amount_in_max=tx_amount_in_max,
                            first_swap=True,
                            last_swap=True,
                        )

                        self.simulated_pool_states.append((v3_pool, v3_sim_result))

                        amount_deposited = max(
                            v3_sim_result.amount0_delta,
                            v3_sim_result.amount1_delta,
                        )

                        if amount_deposited > tx_amount_in_max:
                            raise InsufficientInput(
                                minimum=amount_deposited,
                                deposited=tx_amount_in_max,
                            )

                    case "exactOutput":
                        match func_params["params"], len(func_params["params"]):
                            case dict(), 4 | 5:
                                tx_path = func_params["params"]["path"]
                                tx_recipient = func_params["params"]["recipient"]
                                tx_deadline = func_params["params"].get("deadline")
                                tx_amount_out = func_params["params"]["amountOut"]
                                tx_amount_in_max = func_params["params"]["amountInMaximum"]
                            case tuple(), 5:
                                # Decode with ISwapRouter ABI
                                # https://github.com/Uniswap/v3-periphery/blob/main/contracts/interfaces/ISwapRouter.sol
                                (
                                    tx_path,
                                    tx_recipient,
                                    tx_deadline,
                                    tx_amount_out,
                                    tx_amount_in_max,
                                ) = func_params["params"]
                            case tuple(), 4:
                                # Decode with IV3SwapRouter ABI (aka Router2)
                                # https://github.com/Uniswap/swap-router-contracts/blob/main/contracts/interfaces/IV3SwapRouter.sol
                                (
                                    tx_path,
                                    tx_recipient,
                                    tx_amount_out,
                                    tx_amount_in_max,
                                ) = func_params["params"]
                                tx_deadline = None
                            case _:
                                raise DegenbotValueError(
                                    message=f"Could not identify function parameters. Got {(func_params['params'])}"  # noqa:E501
                                )

                        if tx_deadline:
                            self._raise_if_past_deadline(
                                tx_deadline,
                                block_number=block_number,
                            )

                        tx_path_decoded = decode_v3_path(tx_path)

                        if not self.silent:
                            logger.info(f" • path = {tx_path_decoded}")
                            logger.info(f" • recipient = {tx_recipient}")
                            if tx_deadline is not None:
                                logger.info(f" • deadline = {tx_deadline}")
                            logger.info(f" • amountOut = {tx_amount_out}")
                            logger.info(f" • amountInMaximum = {tx_amount_in_max}")

                        # an exact output path is encoded in REVERSE order,
                        # tokenOut is the first position, tokenIn is the second
                        # position. e.g. tokenOut, fee, tokenIn
                        last_token_pos = len(tx_path_decoded) - 3

                        _amount_in = 0
                        first_swap = False

                        for token_pos in range(
                            0,
                            len(tx_path_decoded) - 2,
                            2,
                        ):
                            tx_token_out_address = tx_path_decoded[token_pos]
                            tx_fee = tx_path_decoded[token_pos + 1]
                            tx_token_in_address = tx_path_decoded[token_pos + 2]

                            first_swap = token_pos == last_token_pos
                            last_swap = token_pos == 0

                            if TYPE_CHECKING:
                                assert isinstance(self.v3_pool_manager, UniswapV3PoolManager)
                                assert isinstance(tx_token_out_address, str)
                                assert isinstance(tx_token_in_address, str)
                                assert isinstance(tx_fee, int)

                            v3_pool = self.v3_pool_manager.get_pool_from_tokens_and_fee(
                                token_addresses=(
                                    tx_token_in_address,
                                    tx_token_out_address,
                                ),
                                pool_fee=tx_fee,
                                silent=self.silent,
                            )

                            _recipient = (
                                tx_recipient if last_swap else UniversalRouterSpecialAddress.ROUTER
                            )
                            _token_in = (
                                v3_pool.token0
                                if v3_pool.token0 == tx_token_in_address
                                else v3_pool.token1
                            )
                            _amount_out = tx_amount_out if last_swap else _amount_in
                            _amount_in_max = tx_amount_in_max if first_swap else None

                            _, v3_sim_result = self._simulate_v3_swap_exact_out(
                                pool=v3_pool,
                                recipient=_recipient,
                                token_in=_token_in,
                                amount_out=_amount_out,
                                amount_in_max=_amount_in_max,
                                first_swap=first_swap,
                                last_swap=last_swap,
                            )

                            _amount_in = max(
                                v3_sim_result.amount0_delta,
                                v3_sim_result.amount1_delta,
                            )

                            _amount_out = -min(
                                v3_sim_result.amount0_delta,
                                v3_sim_result.amount1_delta,
                            )

                            # check that the output of each intermediate swap meets
                            # the input for the next swap
                            if not last_swap:
                                # pool states are appended to `future_pool_states`
                                # so the previous swap will be in the last position
                                _, _last_sim_result = self.simulated_pool_states[-1]

                                if TYPE_CHECKING:
                                    assert isinstance(
                                        _last_sim_result,
                                        UniswapV2PoolSimulationResult
                                        | UniswapV2PoolSimulationResult,
                                    )

                                _last_amount_in = max(
                                    _last_sim_result.amount0_delta,
                                    _last_sim_result.amount1_delta,
                                )

                                if _amount_out != _last_amount_in:
                                    raise InsufficientOutput(
                                        minimum=_last_amount_in,
                                        received=_amount_out,
                                    )

                            self.simulated_pool_states.append((v3_pool, v3_sim_result))

                        # V3 Router enforces a maximum input
                        if first_swap:
                            _sim_result: AbstractSimulationResult
                            _, _sim_result = self.simulated_pool_states[-1]

                            amount_deposited = max(
                                _sim_result.amount0_delta,
                                _sim_result.amount1_delta,
                            )

                            if amount_deposited > tx_amount_in_max:
                                raise InsufficientInput(
                                    minimum=amount_deposited,
                                    deposited=tx_amount_in_max,
                                )

                    case "unwrapWETH9":
                        # TODO: if ETH balances are ever needed, handle the
                        # ETH transfer resulting from this function
                        amount_min = func_params["amountMinimum"]
                        wrapped_token_address = WRAPPED_NATIVE_TOKENS[self.chain_id]
                        wrapped_token_balance = self.ledger.token_balance(
                            self.router_address, wrapped_token_address
                        )
                        if wrapped_token_balance < amount_min:
                            raise InsufficientOutput(
                                minimum=amount_min,
                                received=wrapped_token_balance,
                            )

                        self._simulate_unwrap(wrapped_token_address)

                    case "unwrapWETH9WithFee":
                        # TODO: if ETH balances are ever needed, handle the
                        # two ETH transfers resulting from this function
                        _amount_in = func_params["amountMinimum"]
                        _recipient = func_params["recipient"]
                        _fee_bips = func_params["feeBips"]
                        _fee_recipient = func_params["feeRecipient"]

                        wrapped_token_address = WRAPPED_NATIVE_TOKENS[self.chain_id]
                        wrapped_token_balance = self.ledger.token_balance(
                            self.router_address, wrapped_token_address
                        )
                        if wrapped_token_balance < _amount_in:
                            raise InsufficientOutput(
                                minimum=_amount_in,
                                received=wrapped_token_balance,
                            )

                        self._simulate_unwrap(wrapped_token_address)

                    case "sweepToken":
                        """
                        This function transfers the current token balance
                        held by the contract to `recipient`
                        """

                        tx_token_address = func_params["token"]
                        tx_amount_out_minimum = func_params["amountMinimum"]
                        tx_recipient = func_params.get("recipient")

                        # Router2 ABI omits `recipient`, always uses
                        # `msg.sender`
                        if tx_recipient is None:
                            tx_recipient = self.sender
                        else:
                            self.recipients.add(tx_recipient)

                        _balance = self.ledger.token_balance(self.router_address, tx_token_address)

                        if _balance < tx_amount_out_minimum:
                            raise InsufficientOutput(
                                minimum=tx_amount_out_minimum,
                                received=_balance,
                            )

                        self._simulate_sweep(tx_token_address, tx_recipient)

                    case "wrapETH":
                        _wrapped_token_amount = func_params["value"]
                        _wrapped_token_address = WRAPPED_NATIVE_TOKENS[self.chain_id]
                        self.ledger.adjust(
                            self.router_address,
                            _wrapped_token_address,
                            _wrapped_token_amount,
                        )

            # Wrap exceptions from pool helper simulation attempts
            except LiquidityPoolError as exc:
                raise TransactionError(message=f"Simulation failed: {exc}") from exc

        def _process_uniswap_universal_router_transaction() -> None:
            logger.debug(f"{func_name}: {self.hash.to_0x_hex()=}")

            if func_name != "execute":
                raise DegenbotValueError(
                    message=f"UNHANDLED UNIVERSAL ROUTER FUNCTION: {func_name}"
                )

            try:
                with contextlib.suppress(KeyError):
                    tx_deadline = func_params["deadline"]
                    self._raise_if_past_deadline(
                        tx_deadline,
                        block_number=block_number,
                    )

                tx_commands = func_params["commands"]
                tx_inputs = func_params["inputs"]

                for tx_command, tx_input in zip(tx_commands, tx_inputs, strict=False):
                    _process_universal_router_command(tx_command, tx_input)

            # bugfix: prevents nested multicalls from spamming exception
            # message.
            # e.g. 'Simulation failed: Simulation failed: {error}'
            except TransactionError:
                raise
            except DegenbotError as e:
                raise TransactionError(message=f"Simulation failed: {e}") from e

        if func_name in v2_functions:
            _process_uniswap_v2_transaction()
        elif func_name in v3_functions:
            _process_uniswap_v3_transaction()
        elif func_name in universal_router_functions:
            _process_uniswap_universal_router_transaction()
        elif func_name in unhandled_functions:
            raise TransactionError(
                message=f"Aborting simulation involving un-implemented function: {func_name}"
            )
        elif func_name in no_op_functions:
            logger.debug(f"NON-OP: {func_name}")
        else:
            raise TransactionError(
                message=f"Aborting simulation involving unknown function: {func_name}"
            )

    def simulate(
        self,
        *,
        silent: bool = False,
        state_block: BlockNumber | int | None = None,
    ) -> list[
        tuple[UniswapV2Pool, UniswapV2PoolSimulationResult]
        | tuple[UniswapV3Pool, UniswapV3PoolSimulationResult]
    ]:
        """
        Execute a simulation of a transaction, using the attributes
        stored in the constructor.

        Defers simulation to the `_simulate` method, which may recurse
        as needed for nested multicalls.

        Performs a final accounting check of addresses in `self.ledger`
        ledger, excluding the `msg.sender` and `recipient` addresses.
        """

        self.silent = silent

        self.simulated_pool_states: list[
            tuple[UniswapV2Pool, UniswapV2PoolSimulationResult]
            | tuple[UniswapV3Pool, UniswapV3PoolSimulationResult]
        ] = []

        self._simulate(
            self.func_name,
            self.func_params,
            block_number=(
                connection_manager.get_web3(self.chain_id).eth.get_block_number()
                if state_block is None
                else state_block
            ),
        )

        if self.router_address in self.ledger.balances:
            raise LeftoverRouterBalance(balances=self.ledger.balances[self.router_address])

        return self.simulated_pool_states



import dataclasses
from bisect import bisect_left
from collections.abc import Sequence
from enum import Enum
from fractions import Fraction
from threading import Lock
from typing import TYPE_CHECKING, Any, Final, cast
from weakref import WeakSet

import eth_abi.abi
from eth_abi.exceptions import DecodingError
from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3
from web3.exceptions import ContractLogicError
from web3.types import BlockIdentifier

from degenbot import connection_manager, get_checksum_address
from degenbot.constants import MAX_INT256, MIN_INT256, ZERO_ADDRESS
from degenbot.erc20 import Erc20Token, Erc20TokenManager
from degenbot.exceptions import DegenbotValueError
from degenbot.exceptions.evm import EVMRevertError
from degenbot.exceptions.liquidity_pool import (
    ExternalUpdateError,
    IncompleteSwap,
    LateUpdateError,
    LiquidityMapWordMissing,
    LiquidityPoolError,
    NoPoolStateAvailable,
    PossibleInaccurateResult,
)
from degenbot.functions import encode_function_calldata, raw_call
from degenbot.logging import logger
from degenbot.registry import pool_registry
from degenbot.types.abstract import AbstractArbitrage, AbstractLiquidityPool
from degenbot.types.aliases import BlockNumber, ChainId
from degenbot.types.concrete import (
    AbstractPublisherMessage,
    BoundedCache,
    Publisher,
    PublisherMixin,
    Subscriber,
)
from degenbot.uniswap.types import UniswapPoolSwapVector
from degenbot.uniswap.v3_functions import (
    exchange_rate_from_sqrt_price_x96,
    get_tick_word_and_bit_position,
)
from degenbot.uniswap.v3_types import (
    BitmapWord,
    Liquidity,
    LiquidityGross,
    LiquidityNet,
    Pip,
    SqrtPriceX96,
    Tick,
)
from degenbot.uniswap.v4_libraries.swap_math import (
    MAX_SWAP_FEE,
    compute_swap_step,
    get_sqrt_price_target,
)
from degenbot.uniswap.v4_libraries.tick_bitmap import (
    flip_tick,
    gen_ticks,
    next_initialized_tick_within_one_word,
)
from degenbot.uniswap.v4_libraries.tick_math import (
    MAX_SQRT_PRICE,
    MAX_TICK,
    MIN_SQRT_PRICE,
    MIN_TICK,
    get_sqrt_price_at_tick,
    get_tick_at_sqrt_price,
)
from degenbot.uniswap.v4_types import (
    FeeToProtocol,
    InitializedTickMap,
    LiquidityMap,
    SwapFee,
    UniswapV4BitmapAtWord,
    UniswapV4LiquidityAtTick,
    UniswapV4PoolExternalUpdate,
    UniswapV4PoolKey,
    UniswapV4PoolLiquidityMappingUpdate,
    UniswapV4PoolState,
    UniswapV4PoolStateUpdated,
)


@dataclasses.dataclass(slots=True)
class SwapResult:
    sqrt_price_x96: int
    tick: int
    liquidity: int


@dataclasses.dataclass(slots=True, frozen=True)
class SwapDelta:
    currency0: int
    currency1: int

    @property
    def amount_in(self) -> int:
        "The deposited token amount."
        return -min(self.currency0, self.currency1)

    @property
    def amount_out(self) -> int:
        "The withdrawn token amount."
        return max(self.currency0, self.currency1)


@dataclasses.dataclass(slots=True, frozen=True)
class LiquidityRangeCacheKey:
    """
    Cache key for liquidity range swap calculations.
    """

    exact_input: bool
    tick_lower: int
    tick_upper: int
    liquidity: int
    zero_for_one: bool
    price_start: SqrtPriceX96
    price_end: SqrtPriceX96


@dataclasses.dataclass(slots=True, frozen=True)
class LiquidityRangeCacheValue:
    """
    Cached result of a complete liquidity range consumption.
    """

    amount_in: int
    amount_out: int
    fee_amount: int
    price_end: SqrtPriceX96
    amount_required: int  # Total amount needed to consume this range


@dataclasses.dataclass(slots=True, frozen=True)
class ProtocolFee:
    zero_for_one: int
    one_for_zero: int


@dataclasses.dataclass(slots=True, frozen=True)
class Slot0:
    sqrt_price_x96: int
    tick: int
    protocol_fee: ProtocolFee
    lp_fee: int


PIPS_DENOMINATOR = 1_000_000
NATIVE_CURRENCY_ADDRESS = ZERO_ADDRESS


class Hooks(Enum):
    # ref: https://github.com/Uniswap/v4-core/blob/main/src/libraries/Hooks.sol
    BEFORE_INITIALIZE = 1 << 13
    AFTER_INITIALIZE = 1 << 12
    BEFORE_ADD_LIQUIDITY = 1 << 11
    AFTER_ADD_LIQUIDITY = 1 << 10
    BEFORE_REMOVE_LIQUIDITY = 1 << 9
    AFTER_REMOVE_LIQUIDITY = 1 << 8
    BEFORE_SWAP = 1 << 7
    AFTER_SWAP = 1 << 6
    BEFORE_DONATE = 1 << 5
    AFTER_DONATE = 1 << 4
    BEFORE_SWAP_RETURNS_DELTA = 1 << 3
    AFTER_SWAP_RETURNS_DELTA = 1 << 2
    AFTER_ADD_LIQUIDITY_RETURNS_DELTA = 1 << 1
    AFTER_REMOVE_LIQUIDITY_RETURNS_DELTA = 1 << 0


class UniswapV4Pool(PublisherMixin, AbstractLiquidityPool):
    _state_cache: BoundedCache[BlockNumber, UniswapV4PoolState]

    SLOT0_STRUCT_TYPES = (
        "uint160",  # sqrtPriceX96
        "int24",  # tick
        "uint24",  # protocolFee
        "uint24",  # lpFee
    )
    TICK_LIQUIDITY_STRUCT_TYPES = (
        "uint128",  # liquidityGross
        "int128",  # liquidityNet
    )

    FEE_DENOMINATOR = 1_000_000

    @dataclasses.dataclass(slots=True)
    class SwapState:
        amount_specified_remaining: int
        amount_calculated: int
        sqrt_price_x96: int
        tick: int
        liquidity: int

    @dataclasses.dataclass(slots=True)
    class StepComputations:
        sqrt_price_start_x96: int = 0
        sqrt_price_next_x96: int = 0
        tick_next: int = 0
        initialized: bool = False
        amount_in: int = 0
        amount_out: int = 0
        fee_amount: int = 0
        fee_growth_global_x128: int | None = None  # unused

    def __init__(
        self,
        *,
        pool_id: str,
        pool_manager_address: str,
        state_view_address: str,
        tokens: Sequence[str],
        fee: Pip,
        tick_spacing: int,
        hook_address: str | None = None,
        chain_id: ChainId | None = None,
        tick_data: dict[Tick, dict[str, Any] | UniswapV4LiquidityAtTick] | None = None,
        tick_bitmap: dict[BitmapWord, dict[str, Any] | UniswapV4BitmapAtWord] | None = None,
        state_block: BlockNumber | int | None = None,
        silent: bool = False,
        state_cache_depth: int = 8,
    ) -> None:
        self._chain_id: Final[int] = (
            chain_id if chain_id is not None else connection_manager.default_chain_id
        )
        w3 = connection_manager.get_web3(self.chain_id)
        state_block = state_block if state_block is not None else w3.eth.block_number
        self._initial_state_block = state_block
        self._state_view_address = get_checksum_address(state_view_address)

        assert len(tokens) == 2  # noqa: PLR2004
        tokens = [token.lower() for token in tokens]
        token_manager = Erc20TokenManager(chain_id=self.chain_id)
        self.token0: Final[Erc20Token] = token_manager.get_erc20token(
            address=min(tokens),
            silent=silent,
        )
        self.token1: Final[Erc20Token] = token_manager.get_erc20token(
            address=max(tokens),
            silent=silent,
        )

        self.hook_address = (
            get_checksum_address(hook_address) if hook_address is not None else ZERO_ADDRESS
        )

        self.active_hooks: set[Hooks] = {
            hook for hook in Hooks if int(self.hook_address, 16) & hook.value != 0
        }

        # Construct the PoolKey
        self._pool_key: Final[UniswapV4PoolKey] = UniswapV4PoolKey(
            currency0=self.token0.address,
            currency1=self.token1.address,
            fee=fee,
            tick_spacing=tick_spacing,
            hooks=self.hook_address,
        )

        self._pool_manager_address = get_checksum_address(pool_manager_address)
        self._pool_id: Final[HexBytes] = HexBytes(pool_id)
        self.name = f"{self.token0}-{self.token1} ({self.__class__.__name__}, id={self.pool_id.to_0x_hex()})"

        try:
            _slot0, _liquidity = self._get_state_values(w3=w3, state_block=state_block)
            _sqrt_price_x96 = _slot0.sqrt_price_x96
            _tick = _slot0.tick
            self.lp_fee = _slot0.lp_fee
            self.protocol_fee = _slot0.protocol_fee
        except (ContractLogicError, DecodingError) as exc:
            # Contracts differ slightly across Uniswap V3 forks, so decoding may fail. Catch this
            # here and raise as a pool-specific exception
            raise LiquidityPoolError(message="Could not decode contract data") from exc

        assert self.pool_id == (
            calculated_id := Web3.keccak(
                eth_abi.abi.encode(
                    types=["address", "address", "uint24", "int24", "address"],
                    args=[
                        self.pool_key.currency0,
                        self.pool_key.currency1,
                        self.pool_key.fee,
                        self.pool_key.tick_spacing,
                        self.pool_key.hooks,
                    ],
                )
            )
        ), (
            f"Supplied pool ID {self.pool_id.to_0x_hex()} does not match calculated ID {calculated_id.to_0x_hex()}, {self.pool_key=}"  # noqa
        )

        # If liquidity info was not provided, treat the mapping as sparse
        self.sparse_liquidity_map = tick_bitmap is None or tick_data is None

        _tick_bitmap = {}
        _tick_data = {}

        if tick_bitmap is not None:
            # transform dict to UniswapV4BitmapAtWord
            _tick_bitmap.update(
                {
                    int(word): (
                        UniswapV4BitmapAtWord(**bitmap_at_word)
                        if not isinstance(
                            bitmap_at_word,
                            UniswapV4BitmapAtWord,
                        )
                        else bitmap_at_word
                    )
                    for word, bitmap_at_word in tick_bitmap.items()
                }
            )

        if tick_data is not None:
            _tick_data.update(
                {
                    int(tick): (
                        # transform dict to UniswapV4LiquidityAtTick
                        UniswapV4LiquidityAtTick(**liquidity_at_tick)
                        if not isinstance(
                            liquidity_at_tick,
                            UniswapV4LiquidityAtTick,
                        )
                        else liquidity_at_tick
                    )
                    for tick, liquidity_at_tick in tick_data.items()
                }
            )

        if tick_bitmap is None and tick_data is None:
            word, _ = get_tick_word_and_bit_position(tick=_tick, tick_spacing=self.tick_spacing)
            self._fetch_and_populate_initialized_ticks(
                word_position=word,
                tick_bitmap=_tick_bitmap,
                tick_data=_tick_data,
                block_number=state_block,
            )

        self._state = UniswapV4PoolState(
            id=self.pool_id,
            address=self._pool_manager_address,
            liquidity=_liquidity,
            sqrt_price_x96=_sqrt_price_x96,
            tick=_tick,
            tick_bitmap=_tick_bitmap,
            tick_data=_tick_data,
            block=state_block,
        )
        self._state_cache = BoundedCache(max_items=state_cache_depth)
        self._state_cache[self.update_block] = self.state
        self._state_lock = Lock()

        # Liquidity range consumption cache
        self._swap_step_cache: dict[LiquidityRangeCacheKey, LiquidityRangeCacheValue] = {}
        self._swap_step_cache_lock = Lock()

        pool_registry.add(
            pool=self,
            chain_id=self.chain_id,
            pool_address=self.address,
            pool_id=self.pool_id,
        )

        self._subscribers: WeakSet[Subscriber] = WeakSet()

        if not silent:  # pragma: no branch
            logger.info(self.name)
            logger.info(f"• ID: {self.pool_id.to_0x_hex()}")
            logger.info(f"• Token 0: {self.token0}")
            logger.info(f"• Token 1: {self.token1}")
            logger.info(f"• Liquidity: {self.liquidity}")
            logger.info(f"• SqrtPrice: {self.sqrt_price_x96}")
            logger.info(f"• Tick: {self.tick}")

    def __eq__(self, other: object) -> bool:
        if isinstance(other, type(self)):
            return self.address == other.address and self.pool_id == other.pool_id
        return super().__eq__(other)

    def __hash__(self) -> int:
        return hash(HexBytes(self.address) + self.pool_id)

    def __getstate__(self) -> dict[str, Any]:
        # Remove objects that cannot be pickled and are unnecessary to perform
        # the calculation
        copied_attributes: set[str] = set()
        dropped_attributes = (
            "_state_cache",
            "_state_lock",
            "_subscribers",
            "_swap_step_cache_lock",
        )

        with self._state_lock:
            return {
                k: (v.copy() if k in copied_attributes else v)
                for k, v in self.__dict__.items()
                if k not in dropped_attributes
            }

    def __setstate__(self, state: dict[str, Any]) -> None:
        state["_state_lock"] = Lock()
        state["_swap_step_cache_lock"] = Lock()
        self.__dict__ = state

    def __repr__(self) -> str:  # pragma: no cover
        return f"{self.__class__.__name__}(pool_id={self.pool_id.to_0x_hex()},  token0={self.token0}, token1={self.token1}, fee={self.fee}, tick spacing={self.tick_spacing})"  # noqa:E501

    def __str__(self) -> str:
        return self.name

    def _fetch_and_populate_initialized_ticks(
        self,
        word_position: int,
        tick_bitmap: InitializedTickMap,
        tick_data: LiquidityMap,
        block_number: BlockNumber | None = None,
    ) -> None:
        """
        Update the supplied tick bitmap with initialized tick values within a specified word
        position. A word is divided into 256 ticks, spaced at a fixed interval.
        """

        w3 = connection_manager.get_web3(self.chain_id)

        if block_number is None:
            block_number = w3.eth.get_block_number()

        _tick_bitmap: int = 0
        _tick_data: list[tuple[Tick, LiquidityGross, LiquidityNet]] = []
        _tick_bitmap = self.get_tick_bitmap_at_word(
            w3=w3,
            word_position=word_position,
            block_identifier=block_number,
        )
        if _tick_bitmap != 0:
            _tick_data = self.get_populated_ticks_in_word(
                w3=w3,
                word_position=word_position,
                block_identifier=block_number,
            )

        tick_bitmap[word_position] = UniswapV4BitmapAtWord(
            bitmap=_tick_bitmap,
            block=block_number,
        )
        for tick, liquidity_gross, liquidity_net in _tick_data:
            tick_data[tick] = UniswapV4LiquidityAtTick(
                liquidity_net=liquidity_net,
                liquidity_gross=liquidity_gross,
                block=block_number,
            )

    def _get_state_values(
        self,
        w3: Web3,
        state_block: BlockNumber,
    ) -> tuple[Slot0, Liquidity]:
        with w3.batch_requests() as batch:
            batch.add(
                # This call uses a specific block so the mutable state values are consistent
                w3.eth.call(
                    transaction={
                        "to": self._state_view_address,
                        "data": encode_function_calldata(
                            function_prototype="getSlot0(bytes32)",
                            function_arguments=[self.pool_id],
                        ),
                    },
                    block_identifier=state_block,
                )
            )
            batch.add(
                # This call uses a specific block so the mutable state values are consistent
                w3.eth.call(
                    transaction={
                        "to": self._state_view_address,
                        "data": encode_function_calldata(
                            function_prototype="getLiquidity(bytes32)",
                            function_arguments=[self.pool_id],
                        ),
                    },
                    block_identifier=state_block,
                )
            )

            slot0_result, liquidity_result = batch.execute()

        protocol_fee: int
        price, tick, protocol_fee, lp_fee = eth_abi.abi.decode(
            types=self.SLOT0_STRUCT_TYPES, data=cast("HexBytes", slot0_result)
        )
        liquidity: int
        (liquidity,) = eth_abi.abi.decode(
            types=["uint256"], data=cast("HexBytes", liquidity_result)
        )

        # Extract the two fees from the packed protocol fee
        # ref: https://github.com/Uniswap/v4-core/blob/main/src/types/Slot0.sol
        protocol_fee_as_bytes = (protocol_fee).to_bytes(length=6, byteorder="big")
        protocol_fee_one_to_zero = int.from_bytes(protocol_fee_as_bytes[:3], byteorder="big")
        protocol_fee_zero_to_one = int.from_bytes(protocol_fee_as_bytes[3:6], byteorder="big")

        return (
            Slot0(
                sqrt_price_x96=cast("int", price),
                tick=cast("int", tick),
                protocol_fee=ProtocolFee(
                    one_for_zero=protocol_fee_one_to_zero,
                    zero_for_one=protocol_fee_zero_to_one,
                ),
                lp_fee=cast("int", lp_fee),
            ),
            liquidity,
        )

    def _calculate_swap_fee(
        self,
        protocol_fee: int,
        lp_fee: int,
    ) -> SwapFee:
        protocol_fee &= 0xFFF
        lp_fee &= 0xFFFFFF
        numerator = protocol_fee * lp_fee
        return (protocol_fee + lp_fee) - (numerator // PIPS_DENOMINATOR)

    def _calculate_swap(
        self,
        *,
        zero_for_one: bool,
        amount_specified: int,
        sqrt_price_x96_limit: int,
        override_state: UniswapV4PoolState | None = None,
    ) -> tuple[SwapDelta, FeeToProtocol, SwapFee, SwapResult]:
        """
        This function is ported and adapted from the swap() function implemented by the Pool.sol
        library contract.

        ref: https://github.com/Uniswap/v4-core/blob/main/src/libraries/Pool.sol

        Returns a tuple with amounts and final pool state values for a successful swap:
        (amount0, amount1, sqrt_price_x96, liquidity, tick)

        A positive amount indicates the quantity available for withdrawal by the swapper,
        and a negative amount indicates the deposit required.

        This method will fetch missing liquidity data as needed, but it will be discarded to avoid
        race conditions.
        """

        if override_state is not None:
            liquidity_start = override_state.liquidity
            sqrt_price_x96_start = override_state.sqrt_price_x96
            tick_start = override_state.tick
            tick_bitmap_temp = override_state.tick_bitmap
            tick_data_temp = override_state.tick_data
        else:
            liquidity_start = self.liquidity
            sqrt_price_x96_start = self.sqrt_price_x96
            tick_start = self.tick
            tick_bitmap_temp = self.tick_bitmap
            tick_data_temp = self.tick_data

        protocol_fee = (
            self.protocol_fee.zero_for_one if zero_for_one else self.protocol_fee.one_for_zero
        )

        assert liquidity_start >= 0

        amount_specified_remaining = amount_specified
        amount_calculated = 0
        result: Final[SwapResult] = SwapResult(
            sqrt_price_x96=sqrt_price_x96_start,
            tick=tick_start,
            liquidity=liquidity_start,
        )

        lp_fee = self.lp_fee
        swap_fee = lp_fee if protocol_fee == 0 else self._calculate_swap_fee(protocol_fee, lp_fee)

        # a swap fee totaling MAX_SWAP_FEE (100%) makes exact output swaps impossible since the
        # input is entirely consumed by the fee
        if swap_fee >= MAX_SWAP_FEE and amount_specified > 0:  # exact output
            raise EVMRevertError(error="InvalidFeeForExactOut")

        # swapFee is the pool's fee in pips (LP fee + protocol fee)
        # when the amount swapped is 0, there is no protocolFee applied and the fee amount paid to
        # the protocol is set to 0
        if amount_specified == 0:
            return (
                SwapDelta(currency0=0, currency1=0),
                0,
                swap_fee,
                result,
            )

        if zero_for_one:
            if sqrt_price_x96_limit >= sqrt_price_x96_start:
                raise EVMRevertError(error="PriceLimitAlreadyExceeded")
            # Swaps can never occur at MIN_TICK, only at MIN_TICK + 1, except at initialization of
            # a pool. Under certain circumstances outlined below, the tick will preemptively reach
            # MIN_TICK without swapping there
            if sqrt_price_x96_limit <= MIN_SQRT_PRICE:
                raise EVMRevertError(error="PriceLimitOutOfBounds")
        else:
            if sqrt_price_x96_limit <= sqrt_price_x96_start:
                raise EVMRevertError(error="PriceLimitAlreadyExceeded")
            if sqrt_price_x96_limit >= MAX_SQRT_PRICE:
                raise EVMRevertError(error="PriceLimitOutOfBounds")

        step = self.StepComputations()

        if not self.sparse_liquidity_map:
            # The liquidity mapping is complete. Optimize loop by building a generator that yields
            # ticks and initialization status along the swap path
            ticks_along_swap_path = gen_ticks(
                tick_data=tick_data_temp,
                starting_tick=tick_start,
                tick_spacing=self.tick_spacing,
                less_than_or_equal=zero_for_one,
            )

        while not (
            amount_specified_remaining == 0 or result.sqrt_price_x96 == sqrt_price_x96_limit
        ):
            step.sqrt_price_start_x96 = result.sqrt_price_x96

            if not self.sparse_liquidity_map:
                step.tick_next, step.initialized = next(ticks_along_swap_path)
            else:
                try:
                    step.tick_next, step.initialized = next_initialized_tick_within_one_word(
                        tick_bitmap=tick_bitmap_temp,
                        tick_data=tick_data_temp,
                        tick=result.tick,
                        tick_spacing=self.tick_spacing,
                        less_than_or_equal=zero_for_one,
                    )
                except LiquidityMapWordMissing as exc:
                    missing_word = exc.word
                    self._fetch_and_populate_initialized_ticks(
                        word_position=missing_word,
                        tick_bitmap=tick_bitmap_temp,
                        tick_data=tick_data_temp,
                        block_number=self.update_block,
                    )
                    continue

            # Ensure that we do not overshoot the min/max tick, as the tick bitmap is not aware of
            # these bounds
            step.tick_next = (
                max(MIN_TICK, step.tick_next)  # descending ticks
                if zero_for_one
                else min(MAX_TICK, step.tick_next)  # ascending ticks
            )
            step.sqrt_price_next_x96 = get_sqrt_price_at_tick(step.tick_next)

            # Determine the current liquidity range boundaries
            if zero_for_one:
                tick_lower, tick_upper = (
                    step.tick_next,
                    step.tick_next + self.tick_spacing,
                )
            else:
                tick_lower, tick_upper = (
                    step.tick_next - self.tick_spacing,
                    step.tick_next,
                )
            assert tick_lower < tick_upper, f"{tick_lower} should be < {tick_upper}"

            exact_input = amount_specified < 0

            cached_result = self._swap_step_cache.get(
                LiquidityRangeCacheKey(
                    exact_input=exact_input,
                    tick_lower=tick_lower,
                    tick_upper=tick_upper,
                    liquidity=result.liquidity,
                    zero_for_one=zero_for_one,
                    price_start=step.sqrt_price_start_x96,
                    price_end=step.sqrt_price_next_x96,
                )
            )

            if cached_result and abs(amount_specified_remaining) >= cached_result.amount_required:
                step.amount_in = cached_result.amount_in
                step.amount_out = cached_result.amount_out
                step.fee_amount = cached_result.fee_amount
                result.sqrt_price_x96 = cached_result.price_end
            else:
                # compute values to swap to the target tick, price limit, or point where
                # input/output amount is exhausted
                result.sqrt_price_x96, step.amount_in, step.amount_out, step.fee_amount = (
                    compute_swap_step(
                        sqrt_ratio_x96_current=result.sqrt_price_x96,
                        sqrt_ratio_x96_target=get_sqrt_price_target(
                            zero_for_one=zero_for_one,
                            sqrt_price_next_x96=step.sqrt_price_next_x96,
                            sqrt_price_limit_x96=sqrt_price_x96_limit,
                        ),
                        liquidity=result.liquidity,
                        amount_remaining=amount_specified_remaining,
                        fee_pips=swap_fee,
                    )
                )

            if exact_input:
                if not (MIN_INT256 <= step.amount_in + step.fee_amount <= MAX_INT256):
                    raise EVMRevertError(error="SafeCastOverflow")
                if not (MIN_INT256 <= step.amount_out <= MAX_INT256):
                    raise EVMRevertError(error="SafeCastOverflow")
                amount_specified_remaining += step.amount_in + step.fee_amount
                amount_calculated += step.amount_out
            else:
                if not (MIN_INT256 <= step.amount_out <= MAX_INT256):
                    raise EVMRevertError(error="SafeCastOverflow")
                if not (MIN_INT256 <= step.amount_in + step.fee_amount <= MAX_INT256):
                    raise EVMRevertError(error="SafeCastOverflow")
                amount_specified_remaining -= step.amount_out
                amount_calculated -= step.amount_in + step.fee_amount

            if protocol_fee > 0:
                # step.amountIn does not include the swap fee, as it's already been taken from it,
                # so add it back to get the total amountIn and use that to calculate the amount of
                # fees owed to the protocol cannot overflow due to limits on the size of protocolFee
                # and params.amountSpecified.
                # This rounds down to favor LPs over the protocol
                delta = (
                    step.fee_amount  # lp fee is 0, so the entire fee is owed to the protocol
                    if (swap_fee == protocol_fee)
                    else (step.amount_in + step.fee_amount) * protocol_fee // PIPS_DENOMINATOR
                )
                # subtract it from the total fee and add it to the protocol fee
                step.fee_amount -= delta

            # Shift tick if we reached the next price, and preemptively decrement for zeroForOne
            # swaps to tickNext - 1. If the swap doesn't continue (if amountRemaining == 0 or
            # sqrtPriceLimit is met), slot0.tick will be 1 less than
            # getTickAtSqrtPrice(slot0.sqrtPrice). This doesn't affect swaps, but donation calls
            # should verify both price and tick to reward the correct LPs.
            if result.sqrt_price_x96 == step.sqrt_price_next_x96:
                # Cache the step if it consumed the liquidity range
                if not cached_result:
                    self._cache_swap_step(
                        tick_lower=tick_lower,
                        tick_upper=tick_upper,
                        liquidity=result.liquidity,
                        zero_for_one=zero_for_one,
                        exact_input=exact_input,
                        amount_in=step.amount_in,
                        amount_out=step.amount_out,
                        fee_amount=step.fee_amount,
                        price_start=step.sqrt_price_start_x96,
                        price_end=result.sqrt_price_x96,
                    )
                # If the tick is initialized, adjust the liquidity range
                if step.initialized:
                    liquidity_net_at_next_tick = tick_data_temp[step.tick_next].liquidity_net
                    result.liquidity += (
                        -liquidity_net_at_next_tick if zero_for_one else liquidity_net_at_next_tick
                    )
                result.tick = step.tick_next - 1 if zero_for_one else step.tick_next
            elif result.sqrt_price_x96 != step.sqrt_price_start_x96:
                # Recompute unless we're on a lower tick boundary (i.e. already transitioned ticks),
                # and haven't moved
                result.tick = get_tick_at_sqrt_price(result.sqrt_price_x96)

            assert result.liquidity >= 0

        if zero_for_one != (amount_specified < 0):
            # currency1 is swapped in
            swap_delta = SwapDelta(
                currency0=amount_calculated,
                currency1=amount_specified - amount_specified_remaining,
            )
        else:
            swap_delta = SwapDelta(
                currency0=amount_specified - amount_specified_remaining,
                currency1=amount_calculated,
            )

        return swap_delta, protocol_fee, swap_fee, result

    def _notify_subscribers(self: Publisher, message: AbstractPublisherMessage) -> None:
        for subscriber in self._subscribers:
            subscriber.notify(publisher=self, message=message)

    def calculate_tokens_in_from_tokens_out(
        self,
        token_out: Erc20Token,
        token_out_quantity: int,
        override_state: UniswapV4PoolState | None = None,
    ) -> int:
        if token_out not in self.tokens:  # pragma: no cover
            raise DegenbotValueError(message="token_out not found!")

        zero_for_one = token_out == self.token1

        try:
            swap_delta, *_ = self._calculate_swap(
                zero_for_one=zero_for_one,
                amount_specified=token_out_quantity,
                sqrt_price_x96_limit=MIN_SQRT_PRICE + 1 if zero_for_one else MAX_SQRT_PRICE - 1,
                override_state=override_state,
            )
        except EVMRevertError as e:  # pragma: no cover
            raise LiquidityPoolError(message=f"Simulated execution reverted: {e}") from e

        assert swap_delta.amount_out <= token_out_quantity

        if conflicting_hooks := (
            self.active_hooks
            & {
                Hooks.AFTER_SWAP,
                Hooks.AFTER_SWAP_RETURNS_DELTA,
                Hooks.BEFORE_SWAP,
                Hooks.BEFORE_SWAP_RETURNS_DELTA,
            }
        ):
            raise PossibleInaccurateResult(
                amount_in=swap_delta.amount_in,
                amount_out=swap_delta.amount_out,
                hooks=conflicting_hooks,
            )

        if swap_delta.amount_out < token_out_quantity:
            raise IncompleteSwap(
                amount_in=swap_delta.amount_in,
                amount_out=swap_delta.amount_out,
            )

        return swap_delta.amount_in

    def calculate_tokens_out_from_tokens_in(
        self,
        token_in: Erc20Token,
        token_in_quantity: int,
        override_state: UniswapV4PoolState | None = None,
    ) -> int:
        if token_in not in self.tokens:  # pragma: no cover
            raise DegenbotValueError(message="token_in not found!")

        zero_for_one = token_in == self.token0

        try:
            swap_delta, *_ = self._calculate_swap(
                zero_for_one=zero_for_one,
                amount_specified=-token_in_quantity,
                sqrt_price_x96_limit=MIN_SQRT_PRICE + 1 if zero_for_one else MAX_SQRT_PRICE - 1,
                override_state=override_state,
            )
        except EVMRevertError as e:  # pragma: no cover
            raise LiquidityPoolError(message=f"Simulated execution reverted: {e}") from e

        assert swap_delta.amount_in <= token_in_quantity

        if conflicting_hooks := (
            self.active_hooks
            & {
                Hooks.AFTER_SWAP,
                Hooks.AFTER_SWAP_RETURNS_DELTA,
                Hooks.BEFORE_SWAP,
                Hooks.BEFORE_SWAP_RETURNS_DELTA,
            }
        ):
            raise PossibleInaccurateResult(
                amount_in=swap_delta.amount_in,
                amount_out=swap_delta.amount_out,
                hooks=conflicting_hooks,
            )

        if swap_delta.amount_in < token_in_quantity:
            raise IncompleteSwap(
                amount_in=swap_delta.amount_in,
                amount_out=swap_delta.amount_out,
            )

        return swap_delta.amount_out

    def get_tick_bitmap_at_word(
        self, w3: Web3, word_position: int, block_identifier: BlockIdentifier
    ) -> int:
        (bitmap_at_word,) = raw_call(
            w3=w3,
            address=self._state_view_address,
            calldata=encode_function_calldata(
                function_prototype="getTickBitmap(bytes32,int16)",
                function_arguments=[self.pool_id, word_position],
            ),
            return_types=["uint256"],
            block_identifier=block_identifier,
        )
        return cast("int", bitmap_at_word)

    def get_populated_ticks_in_word(
        self,
        w3: Web3,
        word_position: int,
        block_identifier: BlockIdentifier,
    ) -> list[tuple[int, int, int]]:
        bitmap_at_word = self.get_tick_bitmap_at_word(
            w3=w3, word_position=word_position, block_identifier=block_identifier
        )

        active_ticks = [
            ((word_position << 8) + i) * self.tick_spacing
            for i in range(256)
            if bitmap_at_word & (1 << i) > 0
        ]

        with w3.batch_requests() as batch:
            for tick in active_ticks:
                batch.add(
                    w3.eth.call(
                        transaction={
                            "to": self._state_view_address,
                            "data": encode_function_calldata(
                                function_prototype="getTickLiquidity(bytes32,int24)",
                                function_arguments=[self.pool_id, tick],
                            ),
                        },
                        block_identifier=block_identifier,
                    )
                )
            results = batch.execute()

        populated_ticks = []
        for tick, result in zip(active_ticks, results, strict=True):
            liquidity_gross, liquidity_net = eth_abi.abi.decode(
                types=self.TICK_LIQUIDITY_STRUCT_TYPES,
                data=cast("HexBytes", result),
            )
            populated_ticks.append((tick, liquidity_gross, liquidity_net))

        return populated_ticks

    @property
    def address(self) -> ChecksumAddress:  # type: ignore[override]
        return self._pool_manager_address

    @property
    def chain_id(self) -> int:
        return self._chain_id

    @property
    def liquidity(self) -> int:
        return self.state.liquidity

    @property
    def pool_id(self) -> HexBytes:
        return self._pool_id

    @property
    def pool_key(self) -> UniswapV4PoolKey:
        return self._pool_key

    @property
    def sqrt_price_x96(self) -> int:
        return self.state.sqrt_price_x96

    @property
    def state(self) -> UniswapV4PoolState:
        return self._state

    @property
    def tick(self) -> int:
        return self.state.tick

    @property
    def tick_bitmap(self) -> InitializedTickMap:
        return self.state.tick_bitmap.copy()

    @property
    def tick_data(self) -> LiquidityMap:
        return self.state.tick_data.copy()

    @property
    def tick_spacing(self) -> int:
        return self.pool_key.tick_spacing

    @property
    def fee(self) -> int:
        return self.pool_key.fee

    @property
    def tokens(self) -> tuple[Erc20Token, Erc20Token]:
        return self.token0, self.token1

    @property
    def update_block(self) -> BlockNumber:
        if TYPE_CHECKING:
            assert self.state.block is not None
        return self.state.block

    def swap_is_viable(
        self,
        state: UniswapV4PoolState,
        vector: UniswapPoolSwapVector,
    ) -> bool:
        if self.sparse_liquidity_map:
            # Liquidity cannot be checked with a sparse mapping, so default to True
            return True

        if state.tick_data == {}:
            # The pool has no liquidity
            return False

        if state.sqrt_price_x96 == 0:
            # The pool is not initialized
            assert state.tick_data == {}, (
                f"Found pool @ {self.address} with liquidity positions, but price=0!"
            )
            return False

        if (vector.zero_for_one is True and state.sqrt_price_x96 <= MIN_SQRT_PRICE + 1) or (
            vector.zero_for_one is False and state.sqrt_price_x96 >= MAX_SQRT_PRICE - 1
        ):
            # The price has reached the min/max price, and the swap would drive it beyond
            # that limit
            return False

        # ----------------------------------------------------------------------------------
        # After this point, at least one liquidity position is assumed
        # ----------------------------------------------------------------------------------
        if vector.zero_for_one:
            # A 0->1 swap will lower the price & tick, so pool viability can be
            # determined by checking for a liquidity position starting below
            # the current price
            return get_sqrt_price_at_tick(min(state.tick_data)) < state.sqrt_price_x96
        # A 1->0 swap will raise the price & tick. Check for a liquidity position
        # above the current price, similar to the above comment.
        return get_sqrt_price_at_tick(max(state.tick_data)) > state.sqrt_price_x96

    def auto_update(
        self,
        block_number: BlockNumber | None = None,
        *,
        silent: bool = True,
    ) -> None:
        """
        Retrieves and records the current slot0 and liquidity state from the pool at the provided
        block number, or the latest block if not provided.

        @dev this method uses a lock to guard state-modifying methods that might cause race
        conditions when used with threads.
        """

        with self._state_lock:
            if block_number is not None and block_number < self.update_block:
                raise LateUpdateError

            state_updated = False

            w3 = connection_manager.get_web3(self.chain_id)
            block_number = block_number if block_number is not None else w3.eth.get_block_number()

            _slot0, _liquidity = self._get_state_values(w3=w3, state_block=block_number)
            _sqrt_price_x96 = _slot0.sqrt_price_x96
            _tick = _slot0.tick
            self.lp_fee = _slot0.lp_fee
            self.protocol_fee = _slot0.protocol_fee

            state = dataclasses.replace(
                self.state,
                liquidity=_liquidity,
                sqrt_price_x96=_sqrt_price_x96,
                tick=_tick,
                block=block_number,
            )

            state_updated = state != self.state
            self._state = state
            self._state_cache[block_number] = state

            if state_updated:
                self._notify_subscribers(
                    message=UniswapV4PoolStateUpdated(state),
                )

            if not silent:  # pragma: no cover
                logger.info(f"Liquidity: {self.liquidity}")
                logger.info(f"SqrtPriceX96: {self.sqrt_price_x96}")
                logger.info(f"Tick: {self.tick}")

    def external_update(
        self,
        update: UniswapV4PoolExternalUpdate,
    ) -> bool:
        """
        Process a `UniswapV3PoolExternalUpdate` with one or more of the following update types:
            - `block_number`: int
            - `tick`: int
            - `liquidity`: int
            - `sqrt_price_x96`: int

        `block_number` is validated against the most recently recorded block prior to recording any
        changes.

        Returns a bool indicating whether any updated state value was recorded.

        @dev This method uses a lock to guard state-modifying methods that might cause race
        conditions when used with threads.
        """

        if update.block_number < self.update_block:
            raise ExternalUpdateError(
                message=f"Rejected update for block {update.block_number} in the past, current update block is {self.update_block}"  # noqa:E501
            )

        with self._state_lock:
            state_block = update.block_number

            state = dataclasses.replace(
                self.state,
                liquidity=update.liquidity,
                sqrt_price_x96=update.sqrt_price_x96,
                tick=update.tick,
                block=state_block,
            )

            updated_state = state != self.state

            self._state = state
            self._state_cache[state_block] = state

            if updated_state:
                self._notify_subscribers(
                    message=UniswapV4PoolStateUpdated(state),
                )

            return updated_state

    def update_liquidity_map(
        self,
        update: UniswapV4PoolLiquidityMappingUpdate,
    ) -> None:
        """
        Applies an update to the liquidity map.

        @dev This method uses a lock to guard state-modifying methods that might cause race
        conditions when used with threads.
        """

        if update.liquidity == 0:
            return

        with self._state_lock:
            # Invalidate cache entries for the affected range before applying the update
            self._invalidate_range_cache_for_ticks(update.tick_lower, update.tick_upper)

            state_block = update.block_number

            # The tick bitmap and tick data dictionaries are copies, so they can be freely modified
            # without corrupting states for previous blocks
            _tick_bitmap = self.tick_bitmap
            _tick_data = self.tick_data

            _liquidity = self.liquidity

            assert _liquidity >= 0, (
                f"Starting liquidity violates invariant: pool {self.address} {self.tick=} {self.liquidity=}"  # noqa: E501
            )

            # Adjust in-range liquidity if the modified region includes the active tick.
            # NOTE: This compares the update block to `initial_state_block` so that onchain
            # liquidity updates from blocks prior to the creation of this pool helper can be applied
            # without triggering an inconsistent invariant check. Particularly, the values for
            # `self.tick` and `self.liquidity` may not align with the pool state when these
            # liquidity events occured.
            if (
                update.tick_lower <= self.tick < update.tick_upper
                and state_block > self._initial_state_block
            ):
                _liquidity += update.liquidity
                assert _liquidity >= 0, (
                    f"In-range liquidity adjustment violated invariant: pool {self.address} {self.tick=} {self.liquidity=} {self.update_block=} {update=}"  # noqa: E501
                )

            for tick in (update.tick_lower, update.tick_upper):
                tick_word, _ = get_tick_word_and_bit_position(tick, self.tick_spacing)

                if self.sparse_liquidity_map and tick_word not in _tick_bitmap:
                    # The liquidity map at the affected word must be complete prior to changing the
                    # status of any tick
                    self._fetch_and_populate_initialized_ticks(
                        word_position=tick_word,
                        tick_bitmap=_tick_bitmap,
                        tick_data=_tick_data,
                        block_number=(
                            # Populate the liquidity data from the previous block
                            state_block - 1
                        ),
                    )

                # Get the liquidity info for this tick. If the mapping is empty at this tick, it is
                # uninitialized and must be flipped in the bitmap and initialized as empty in the
                # mapping
                if tick not in _tick_data:
                    _tick_data[tick] = UniswapV4LiquidityAtTick(
                        liquidity_net=0,
                        liquidity_gross=0,
                        block=state_block,
                    )
                    flip_tick(
                        tick_bitmap=_tick_bitmap,
                        sparse=self.sparse_liquidity_map,
                        tick=tick,
                        tick_spacing=self.tick_spacing,
                        update_block=state_block,
                    )

                current_liquidity_net = _tick_data[tick].liquidity_net
                current_liquidity_gross = _tick_data[tick].liquidity_gross

                new_liquidity_gross = current_liquidity_gross + update.liquidity
                assert new_liquidity_gross >= 0, (
                    f"Negative gross liquidity ({new_liquidity_gross})!"
                )

                if new_liquidity_gross == 0:
                    # Delete tick from the map if there is no remaining liquidity referencing it,
                    # and flip it in the bitmap
                    del _tick_data[tick]
                    flip_tick(
                        tick_bitmap=_tick_bitmap,
                        sparse=self.sparse_liquidity_map,
                        tick=tick,
                        tick_spacing=self.tick_spacing,
                        update_block=state_block,
                    )
                    continue

                # Liquidity positions include the lower tick, but exclude the upper tick.
                if tick == update.tick_lower:
                    new_liquidity_net = current_liquidity_net + update.liquidity
                else:
                    new_liquidity_net = current_liquidity_net - update.liquidity

                _tick_data[tick] = UniswapV4LiquidityAtTick(
                    liquidity_net=new_liquidity_net,
                    liquidity_gross=new_liquidity_gross,
                    block=state_block,
                )

            state = dataclasses.replace(
                self.state,
                liquidity=_liquidity,
                tick_data=_tick_data,
                tick_bitmap=_tick_bitmap,
                block=max(self.update_block, state_block),
            )
            self._state = state
            self._state_cache[state_block] = state
            self._notify_subscribers(
                message=UniswapV4PoolStateUpdated(state),
            )

    def get_arbitrage_helpers(self) -> tuple[AbstractArbitrage, ...]:
        return tuple(
            subscriber
            for subscriber in self._subscribers
            if isinstance(subscriber, AbstractArbitrage)
        )

    def get_absolute_price(
        self,
        token: Erc20Token,
        override_state: UniswapV4PoolState | None = None,
    ) -> Fraction:
        """
        Get the absolute price for the given token, expressed in units of the other.
        """

        return 1 / self.get_absolute_exchange_rate(token, override_state=override_state)

    def get_absolute_exchange_rate(
        self,
        token: Erc20Token,
        override_state: UniswapV4PoolState | None = None,
    ) -> Fraction:
        """
        Get the absolute exchange rate for the given token, expressed in terms of a unit amount of
        its paired token.

        e.g. taking the USDC-WETH pool in https://blog.uniswap.org/uniswap-v3-math-primer — the
        WETH/USDC exchange rate is 649004842.70137. Rounding down, this signifies that the smallest
        swap (1 USDC) results in a 649004842 WETH output.

        A V4 pool encodes the token1/token0 exchange rate in `sqrt_price_x96`, so it can be directly
        obtained.
        """

        if token not in self.tokens:
            raise DegenbotValueError(message=f"Unknown token {token}")

        state = self.state if override_state is None else override_state

        return (
            exchange_rate_from_sqrt_price_x96(state.sqrt_price_x96)
            if token == self.token1
            else 1 / exchange_rate_from_sqrt_price_x96(state.sqrt_price_x96)
        )

    def get_nominal_price(
        self,
        token: Erc20Token,
        override_state: UniswapV4PoolState | None = None,
    ) -> Fraction:
        """
        Get the nominal price for the given token, expressed in units of the other, corrected for
        decimal place values.
        """

        return 1 / self.get_nominal_exchange_rate(token, override_state=override_state)

    def get_nominal_exchange_rate(
        self,
        token: Erc20Token,
        override_state: UniswapV4PoolState | None = None,
    ) -> Fraction:
        """
        Get the nominal rate for the given token, expressed in units of the other, corrected for
        decimal place values.
        """

        return self.get_absolute_exchange_rate(token=token, override_state=override_state) * (
            Fraction(10**self.token1.decimals, 10**self.token0.decimals)
            if token == self.token0
            else Fraction(10**self.token0.decimals, 10**self.token1.decimals)
        )

    def discard_states_before_block(
        self,
        block: BlockNumber,
    ) -> None:
        """
        Discard states recorded prior to a target block.
        """

        with self._state_lock:
            known_blocks = sorted(self._state_cache.keys())

            # Finds the index prior to the requested block number
            block_index = bisect_left(known_blocks, block)

            # The earliest known state already meets the criterion, so return early
            if block_index == 0:
                return

            if block_index == len(known_blocks):
                raise NoPoolStateAvailable(block)

            for known_block in known_blocks[:block_index]:
                del self._state_cache[known_block]

    def restore_state_before_block(
        self,
        block: BlockNumber,
    ) -> None:
        """
        Restore the last pool state recorded prior to a target block.

        Use this method to maintain consistent state data following a chain re-organization.
        """

        # Find the index for the most recent pool state PRIOR to the requested
        # block number.
        #
        # e.g. Calling restore_state_before_block(block=104) for a pool with
        # states at blocks 100, 101, 102, 103, 104. `bisect_left()` returns
        # block_index=3, since block 104 is at index=4. The state held at
        # index=3 is for block 103.

        with self._state_lock:
            known_blocks = sorted(self._state_cache.keys())
            block_index = bisect_left(known_blocks, block)

            if block_index == 0:
                raise NoPoolStateAvailable(block)

            # The last known state already meets the criterion, so return early
            if block_index == len(known_blocks):
                return

            # Remove states at and after the specified block
            for known_block in known_blocks[block_index:]:
                del self._state_cache[known_block]

            self._update_block, self._state = list(self._state_cache.items())[-1]

        self._notify_subscribers(
            message=UniswapV4PoolStateUpdated(self.state),
        )

    def _cache_swap_step(
        self,
        *,
        tick_lower: int,
        tick_upper: int,
        liquidity: int,
        zero_for_one: bool,
        exact_input: bool,
        amount_in: int,
        amount_out: int,
        fee_amount: int,
        price_start: SqrtPriceX96,
        price_end: SqrtPriceX96,
    ) -> None:
        """
        Cache the result of a swap step that consumed an entire liquidity range.
        """
        cache_key = LiquidityRangeCacheKey(
            exact_input=exact_input,
            tick_lower=tick_lower,
            tick_upper=tick_upper,
            liquidity=liquidity,
            zero_for_one=zero_for_one,
            price_start=price_start,
            price_end=price_end,
        )

        # Calculate the total amount required to consume this range
        amount_required = (amount_in + fee_amount) if exact_input else amount_out

        cache_value = LiquidityRangeCacheValue(
            amount_in=amount_in,
            amount_out=amount_out,
            fee_amount=fee_amount,
            price_end=price_end,
            amount_required=amount_required,
        )

        with self._swap_step_cache_lock:
            self._swap_step_cache[cache_key] = cache_value

    def _invalidate_range_cache_for_ticks(self, tick_lower: int, tick_upper: int) -> None:
        """
        Invalidate cache entries that overlap with the specified tick range.

        This should be called when liquidity is updated in a range to ensure
        cached calculations remain valid.
        """
        with self._swap_step_cache_lock:
            self._swap_step_cache = {
                key: value
                for key, value in self._swap_step_cache.items()
                # Preserve only the cached values outside of the updated range
                # TODO: determine if one or more inequalities can be <= or >=
                if (tick_lower > key.tick_upper or tick_upper < key.tick_lower)
            }

    def clear_swap_step_cache(self) -> None:
        """
        Clear all cached liquidity range calculations.

        This can be useful for memory management or when you want to ensure
        fresh calculations.
        """
        with self._swap_step_cache_lock:
            self._swap_step_cache.clear()

    def get_range_cache_stats(self) -> dict[str, int]:
        """
        Get statistics about the liquidity range cache.

        Returns:
            Dictionary containing cache size and other metrics.
        """
        with self._swap_step_cache_lock:
            return {
                "cache_size": len(self._swap_step_cache),
            }



from collections.abc import Sequence
from typing import cast

from degenbot.balancer.libraries.constants import ONE
from degenbot.balancer.libraries.fixed_point import div_down, div_up, mul_down
from degenbot.erc20 import Erc20Token

# To simplify Pool logic, all token balances and amounts are normalized to behave as if the token
# had 18 decimals. e.g. When comparing DAI (18 decimals) and USDC (6 decimals), 1 USDC and 1 DAI
# would both be represented as 1e18, whereas without scaling 1 USDC would be represented as 1e6.
# This allows us to not consider differences in token decimals in the internal Pool maths,
# simplifying it greatly.


def _upscale(
    amount: int,
    scaling_factor: int,
) -> int:
    # Upscale rounding wouldn't necessarily always go in the same direction: in a swap for example
    # the balance of token in should be rounded up, and that of token out rounded down. This is the
    # only place where we round in the same direction for all amounts, as the impact of this
    # rounding is expected to be minimal.

    return mul_down(amount, scaling_factor)


def _downscale_down(
    amount: int,
    scaling_factor: int,
) -> int:
    """
    Reverses the `scaling_factor` applied to `amount`, resulting in a smaller or equal value
    depending on whether it needed scaling or not. The result is rounded down.
    """

    return div_down(amount, scaling_factor)


def _downscale_up(
    amount: int,
    scaling_factor: int,
) -> int:
    """
    Reverses the `scaling_factor` applied to `amount`, resulting in a smaller or equal value
    depending on whether it needed scaling or not. The result is rounded up.
    """

    return div_up(amount, scaling_factor)


def _upscale_array(amounts: list[int], scaling_factors: Sequence[int]) -> None:
    """
    Same as `_upscale`, but for an entire array. This function does not return anything, but instead
    *mutates* the `amounts` array.
    """

    amounts = [
        mul_down(amount, factor) for amount, factor in zip(amounts, scaling_factors, strict=True)
    ]


def _downscale_down_array(amounts: list[int], scaling_factors: list[int]) -> None:
    """
    Same as `_downscale_down`, but for an entire array. This function does not return anything, but
    instead *mutates* the `amounts` array.
    """

    amounts = [
        div_down(amount, factor) for amount, factor in zip(amounts, scaling_factors, strict=True)
    ]


def _downscale_up_array(amounts: list[int], scaling_factors: list[int]) -> None:
    """
    Same as `_downscale_up`, but for an entire array. This function does not return anything, but
    instead *mutates* the `amounts` array.
    """

    amounts = [
        div_up(amount, factor) for amount, factor in zip(amounts, scaling_factors, strict=True)
    ]


def _compute_scaling_factor(token: Erc20Token) -> int:
    # Tokens that don't implement the `decimals` method are not supported.
    token_decimals = token.decimals

    # Tokens with more than 18 decimals are not supported.
    decimals_difference = 18 - token_decimals
    return cast("int", ONE * 10**decimals_difference)
